"""Interpolation of a sound field taking physical properties of sound into account.

The estimation methods are written generally to allow for any kernel function to be given as argument, 
and then functions implementing the kernel functions associated with the papers below are provided. 

A number of kernels are implemented

- Gaussian kernel

- Diffuse kernel in 2D [itoFeedforward2019] and 3D [uenoKernel2018]

- Directional kernel in 3D [uenoDirectionally2021, koyamaSpatial2021]

- Reciprocal kernel for RIR estimation [ribeiroKernel2020]

References
----------
[uenoKernel2018] N. Ueno, S. Koyama, and H. Saruwatari, “Kernel ridge regression with constraint of Helmholtz equation for sound field interpolation,” in 2018 16th International Workshop on Acoustic Signal Enhancement (IWAENC), Tokyo, Japan: IEEE, Sep. 2018, pp. 436–440. doi: 10.1109/IWAENC.2018.8521334. `[link] <https://doi.org/10.1109/IWAENC.2018.8521334>`__ \n
[uenoDirectionally2021] N. Ueno, S. Koyama, and H. Saruwatari, “Directionally weighted wave field estimation exploiting prior information on source direction,” IEEE Transactions on Signal Processing, vol. 69, pp. 2383–2395, Apr. 2021, doi: 10.1109/TSP.2021.3070228. `[link] <https://doi.org/10.1109/TSP.2021.3070228>`__ \n
[ribeiroKernel2020] J. G. C. Ribeiro, N. Ueno, S. Koyama, and H. Saruwatari, “Kernel interpolation of acoustic transfer function between regions considering reciprocity,” in 2020 IEEE 11th Sensor Array and Multichannel Signal Processing Workshop (SAM), Jun. 2020, pp. 1–5. doi: 10.1109/SAM48682.2020.9104256. `[link] <https://doi.org/10.1109/SAM48682.2020.9104256>`__ \n
[koyamaSpatial2021] S. Koyama, J. Brunnström, H. Ito, N. Ueno, and H. Saruwatari, “Spatial active noise control based on kernel interpolation of sound field,” IEEE/ACM Transactions on Audio, Speech, and Language Processing, vol. 29, pp. 3052–3063, Aug. 2021, doi: 10.1109/TASLP.2021.3107983. `[link] <https://doi.org/10.1109/TASLP.2021.3107983>`__ \n
[brunnstromVariable2022] J. Brunnström, S. Koyama, and M. Moonen, “Variable span trade-off filter for sound zone control with kernel interpolation weighting,” in ICASSP 2022 - 2022 IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP), May 2022, pp. 1071–1075. doi: 10.1109/ICASSP43922.2022.9746550. `[link] <https://doi.org/10.1109/ICASSP43922.2022.9746550>`__ \n
[itoFeedforward2019] H. Ito, S. Koyama, N. Ueno, and H. Saruwatari, “Feedforward spatial active noise control based on kernel interpolation of sound field,” in IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP), Brighton, United Kingdom: IEEE, May 2019, pp. 511–515. doi: 10.1109/ICASSP.2019.8683067. `[link] <https://doi.org/10.1109/ICASSP.2019.8683067>`__ \n
"""
import numpy as np
import scipy.spatial.distance as distfuncs
import scipy.special as special
import numba as nb

import aspcol.utilities as util
import aspcore.fouriertransform as ft
import aspcore.montecarlo as mc
import aspcore.filterdesign as fd

def kernel_gaussian(points1, points2, scale):
    """
    Guassian kernel, also known as the radial basis function kernel. 

    Parameters
    ----------
    points1 : ndarray of shape (num_points1, spatial_dim)
    points2 : ndarray of shape (num_points2, spatial_dim)
    scale : ndarray of shape (num_scales,)

    Returns
    -------
    ndarray of shape (num_scales, num_points1, num_points2)
    """
    dist_mat = distfuncs.cdist(points1, points2)**2
    return np.exp(-scale[:,None,None]**2 * dist_mat[None,:,:])

def kernel_helmholtz_2d(points1, points2, wave_num):
    """
    Parameters
    ----------
    points1 : ndarray of shape (num_points1, 2)
    points2 : ndarray of shape (num_points2, 2)
    wave_num : ndarray of shape (num_freqs,)

    Returns
    -------
    ndarray of shape (num_freqs, num_points1, num_points2)
    """
    dist_mat = distfuncs.cdist(points1, points2)
    return special.j0(dist_mat[None,:,:] * wave_num[:,None,None])


def kernel_helmholtz_3d_slow(points1, points2, wave_num):
    """
    Identical to kernel_helmholtz_3d, but is not JIT compiled by numba. 
    This is faster if the kernel is only evaluated once, but slower if it is evaluated many times
    
    Parameters
    ----------
    points1 : ndarray of shape (num_points1, 3)
    points2 : ndarray of shape (num_points2, 3)
    wave_num : ndarray of shape (num_freqs,)

    Returns
    -------
    ndarray of shape (num_freqs, num_points1, num_points2)
    """
    distMat = distfuncs.cdist(points1, points2)
    return special.spherical_jn(0, distMat[None,:,:] * wave_num[:,None,None])


@nb.njit
def kernel_helmholtz_3d(points1, points2, wave_num):
    """
    Diffuse kernel for 3D sound field interpolation. Defined in 
    'Spatial active noise control based on kernel interpolation 
    of sound field' by Koyama et al.

    Parameters
    ----------
    points1 : ndarray of shape (num_points1, 3)
    points2 : ndarray of shape (num_points2, 3)
    wave_num : ndarray of shape (num_freqs,)

    Returns
    -------
    ndarray of shape (num_freqs, num_points1, num_points2)
    """
    #distMat = distfuncs.cdist(points1, points2)

    dist_mat = np.sqrt(np.sum((np.expand_dims(points1,1) - np.expand_dims(points2,0))**2, axis=-1))
    return np.sinc(np.expand_dims(dist_mat,0) * wave_num.reshape(-1, 1,1) / np.pi)


def kernel_directional_3d_slow(points1, points2, wave_num, angle, beta):    
    """
    Identical to kernel_directional_3d, but is not JIT compiled by numba. 
    This is faster if the kernel is only evaluated once, but slower if it is evaluated many times

    In addition, this only allows for a single angle to be evaluated at a time.

    Parameters
    ----------
    points1 : ndarray of shape (num_points1, 3)
    points2 : ndarray of shape (num_points2, 3)
    wave_num : ndarray of shape (num_freqs,)
    angle : tuple (theta, phi) defined as in util.spherical2cart
    beta : sets the strength of the directional weighting

    Returns
    -------
    ndarray of shape (num_freqs, num_points1, num_points2)
    """
    rDiff = points1[:,None,:] - points2[None,:,:]
    angleFactor = beta * util.spherical2cart(np.ones((1,1)), np.array(angle)[None,:])[None,None,...]
    posFactor = 1j * wave_num[:,None,None,None] * rDiff[None,...]
    return special.spherical_jn(0, 1j*np.sqrt(np.sum((angleFactor + posFactor)**2, axis=-1)))

@nb.njit
def kernel_directional_3d(points1, points2, wave_num, direction_vec, beta):
    """
    Directionally weighted kernel for 3D sound field interpolation. 
    Defined in 'Spatial active noise control based on kernel interpolation 
    of sound field' by Koyama, Brunnström, Ito, Ueno, Saruwatari.
    
    Parameters
    ----------
    points1 : ndarray of shape (num_points1, 3)
    points2 : ndarray of shape (num_points2, 3)
    wave_num : ndarray of shape (num_freqs,)
    direction_vec : ndarray of shape (num_angles, 3)
        unit vectors describing the arrival direction
    beta : float nonnegative 
        sets the strength of the directional weighting

    Returns
    -------
    ndarray of shape (num_freqs, num_angles, num_points1, num_points2)
    """
    angle_term = 1j * beta * direction_vec.reshape((1,-1,1,1,direction_vec.shape[-1]))
    pos_term = wave_num.reshape((-1,1,1,1,1)) * (points1.reshape((1,1,-1,1,points1.shape[-1])) - points2.reshape((1,1,1,-1,points2.shape[-1])))
    return np.sinc(np.sqrt(np.sum((angle_term - pos_term)**2, axis=-1)) / np.pi)


def kernel_reciprocal_3d(points1, points2, wave_num):
    """
    Reciprocal kernel for room impulse response interpolation. Definition found in
    'Kernel interpolation of acoustic transfer function between regions considering reciprocity'
    by Ribeiro, Ueno, Koyama, Saruwatari.

    Parameters
    ----------
    points1 : 2-tuple of (mic_points1, src_points1)
        where mic_points ndarray of shape (num_mic1, 3)
        and src_points ndarray of shape (num_src1, 3)
    points2 : same type of object as points1, 
        although the ndarray shapes can be different
    wave_num : ndarray of shape (num_freqs,)

    Returns
    -------
    ndarray of shape (num_freqs, num_mic1*num_src1, num_mic2*num_src2)
        When flattened, the index for microphone m, speaker l is m+l*M. i.e.
        the microphone index changes faster. 
    """
    wave_num = wave_num[:,None,None,None,None]
    mic_dist = distfuncs.cdist(points1[0], points2[0])[None,None,:,None,:]
    src_dist = distfuncs.cdist(points1[1], points2[1])[None,:,None,:,None]
    mix_dist1 = distfuncs.cdist(points1[0], points2[1])[None,None,:,:,None]
    mic_dist2 = distfuncs.cdist(points1[1], points2[0])[None,:,None,None,:]

    k_val = 0.5 * (special.spherical_jn(0, wave_num * mic_dist) * \
                        special.spherical_jn(0, wave_num * src_dist)) + \
                        (special.spherical_jn(0, wave_num * mix_dist1) * \
                        special.spherical_jn(0, wave_num * mic_dist2))
    k_val = np.reshape(k_val, k_val.shape[:3]+(-1,))
    k_val = np.reshape(k_val, (k_val.shape[0], -1,k_val.shape[-1]))
    return k_val


def get_kernel_weighting_filter(kernel_func, reg_param, mic_pos, integral_domain, 
                                mc_samples, num_freq, samplerate, c, *args):
    """ 
    Calculates kernel weighting filter A(w) in frequency domain
    see 'Spatial active noise control based on kernel interpolation of sound field' by Koyama et al.     

    Parameters
    ----------
    kernel_func : function
        with calling signature kernel_func(points1, points2, waveNum, *args)
    reg_param : float
    mic_pos : ndarray of shape (num_mics, spatial_dim)
    integral_domain : instance of any Region object found is aspsim package
    mc_samples : int
        how many monte carlo samples to be drawn for integration
    num_freq : int
    samplerate : int
    c : float
        speed of sound
    *args : arguments needed for kernel function except points1, points2, waveNum

    Returns
    -------
    ndarray of shape (num_freq, num_mics, num_mics)
    
    For both diffuse and directional kernel P^H = P, so the hermitian tranpose should not do anything
    It is left in place in case a kernel function in the future changes that identity. 
    """
    freqs = ft.get_real_freqs(num_freq, samplerate)
    wave_num = 2 * np.pi * freqs / c

    def integrable_func(r):
        kappa = kernel_func(r, mic_pos, wave_num, *args)
        kappa = np.transpose(kappa,(0,2,1))
        return kappa.conj()[:,:,None,:] * kappa[:,None,:,:]

    num_mics = mic_pos.shape[0]
    K = kernel_func(mic_pos, mic_pos, wave_num, *args)
    P = np.linalg.pinv(K + reg_param * np.eye(num_mics))

    integral_value = mc.integrate(integrable_func, integral_domain.sample_points, mc_samples, integral_domain.volume)
    weighting_filter = np.transpose(P,(0,2,1)).conj() @ integral_value @ P

    weighting_filter = ft.insert_negative_frequencies(weighting_filter, even=True)
    return weighting_filter



def get_krr_parameters(kernel_func, reg_param, output_arg, data_arg, *args):
    """
    Calculates parameter vector or matrix given a kernel function for Kernel Ridge Regression.
    The returned parameter Z is the optimal interpolation filter from the data points to
    the output points. Apply filter as Z @ y, where y are the labels for data at data_arg positions
    
    Parameters
    ----------
    data_arg : ndarray (num_data_points, data_dim)
    output_arg : ndarray (num_out_points, data_dim)
    kernel_func : function 
        with calling signature kernel_func(output_arg, data_arg, \*args)
        should return ndarray (..., num_out_points, num_data_poins)
    
    Returns
    -------
    params : ndarray (..., num_out_points, num_data_points)
    """
    K = kernel_func(data_arg, data_arg, *args)
    K_reg = K + reg_param * np.eye(K.shape[-1])
    kappa = np.moveaxis(kernel_func(output_arg, data_arg, *args), -1, -2)

    params = np.moveaxis(np.linalg.solve(np.moveaxis(K_reg, -1, -2), kappa), -1, -2)
    return params


def soundfield_interpolation_fir(
    to_points, from_points, ir_len, reg_param, num_freq, spatial_dims, samplerate, c
):
    """
    Convenience function for calculating the time domain causal FIR interpolation filter
    from a set of points to another set of points. 
    
    Parameters
    ----------
    to_points : ndarray of shape (num_to_points, spatial_dims)
    from_points : ndarray of shape (num_from_points, spatial_dims)
    ir_len : int
        length of the impulse response
    reg_param : float
    num_freq : int
    spatial_dims : int
    samplerate : int
    c : float
        speed of sound

    Returns
    -------
    ndarray of shape (ir_len, num_to_points, num_from_points)    
    """
    assert num_freq > ir_len
    freq_filter = soundfield_interpolation(
        to_points, from_points, num_freq, reg_param, spatial_dims, samplerate, c
    )
    ki_filter,_ = fd.fir_from_freqs_window(freq_filter, ir_len)
    return ki_filter

def soundfield_interpolation(
    to_points, from_points, num_freq, reg_param, spatial_dims, samplerate, c
):
    """ Convenience function for calculating the frequency domain interpolation filter
    from a set of points to another set of points. 
    
    Parameters
    ----------
    to_points : ndarray of shape (num_to_points, spatial_dims)
    from_points : ndarray of shape (num_from_points, spatial_dims)
    num_freq : int
    reg_param : float
    spatial_dims : int
    samplerate : int
    c : float
        speed of sound

    Returns
    -------
    ndarray of shape (num_freq, num_to_points, num_from_points)
    """
    if spatial_dims == 3:
        kernel_func = kernel_helmholtz_3d
    elif spatial_dims == 2:
        kernel_func = kernel_helmholtz_2d
    else:
        raise ValueError

    assert num_freq % 2 == 0

    freqs = ft.get_real_freqs(num_freq, samplerate)
    wave_num = 2 * np.pi * freqs / c
    ip_params = get_krr_parameters(kernel_func, reg_param, to_points, from_points, wave_num)
    ip_params = ft.insert_negative_frequencies(ip_params, even=True)
    return ip_params




def analytic_kernel_weighting_disc_2d(error_mic_pos, freq, reg_param, trunc_order, radius, c):
    """
    Analytic solution of the kernel interpolation weighting filter integral
    in the frequency domain for a disc in 2D. See definition in 'Feedforward 
    spatial active noise control based on kernel interpolation of sound field'
    by Ito, Koyama, Ueno, Saruwatari.

    Parameters
    ----------
    error_mic_pos : ndarray of shape (num_mics, spatial_dim)
    freq : float or ndarray of shape (num_freqs,)
    reg_param : float
    trunc_order : int
    radius : float
    c : float

    Returns
    -------
    ndarray of shape (num_freqs, num_mics, num_mics)
    """
    if isinstance(freq, (int, float)):
        freq = np.array([freq])
    if len(freq.shape) == 1:
        freq = freq[:, np.newaxis, np.newaxis]
    wave_number = 2 * np.pi * freq / c
    K = special.j0(wave_number * distfuncs.cdist(error_mic_pos, error_mic_pos))
    P = np.linalg.pinv(K + reg_param * np.eye(K.shape[-1]))
    S = _get_s(trunc_order, wave_number, error_mic_pos)
    gamma = _get_gamma(trunc_order, wave_number, radius)
    A = (
        np.transpose(P.conj(), (0, 2, 1))
        @ np.transpose(S.conj(), (0, 2, 1))
        @ gamma
        @ S
        @ P
    )
    return A

def _get_gamma(maxOrder, k, R):
    matLen = 2 * maxOrder + 1
    diagValues = _small_gamma(np.arange(-maxOrder, maxOrder + 1), k, R)
    gamma = np.zeros((diagValues.shape[0], matLen, matLen))

    gamma[:, np.arange(matLen), np.arange(matLen)] = diagValues
    return gamma

def _small_gamma(mu, k, R):
    Jfunc = special.jv((mu - 1, mu, mu + 1), k * R)
    return np.pi * (R ** 2) * ((Jfunc[:, 1, :] ** 2) - Jfunc[:, 0, :] * Jfunc[:, 2, :])

def _get_s(maxOrder, k, positions):
    r, theta = util.cart2pol(positions[:, 0], positions[:, 1])

    mu = np.arange(-maxOrder, maxOrder + 1)[:, np.newaxis]
    S = special.jv(mu, k * r) * np.exp(theta * mu * (-1j))
    return S


import numpy as np
import scipy.special as special

import aspcore.montecarlo as mc
import aspcore.fouriertransform as ft

import aspcore.utilities as coreutils

import aspcol.sphericalharmonics as sph
import aspcol.utilities as utils
import plot_methods as plm
import aspcol.planewaves as pw

from aspsim.simulator import SimulatorSetup
import aspsim.room.region as region

import matplotlib.pyplot as plt

def show_plane_wave_response():
    freq = 400
    c = 343
    wave_num = 2 * np.pi * freq / c
    num_pos = 128
    direction = np.array([1,1,0])

    posx = np.linspace(-1, 1, num_pos)
    posy = np.linspace(-1, 1, num_pos)
    (posx, posy) = np.meshgrid(posx, posy)
    pos = np.stack((posx.flatten(), posy.flatten(), np.zeros(num_pos**2)), axis=-1)

    p = pw.plane_wave(pos, direction, wave_num)
    plm.image_scatter_freq_response(p, np.array([freq]), pos, dot_size=5)
    plt.show()


def _approx_dirac_planewave_coeffs(direction, width):
    """
    direction : ndarray of shape (3,)
        The direction of delta function. Must be a unit vector
    width : float
        The width of the approximate delta function, Measured as the distance from the direction vector. 
        The decision boundary is therefore a sphere-sphere intersection.
        The smaller the width, the more accuracte, but more prone to numerical errors.
    """

    #proj_len = np.sum(dir_vec * direction[None,:], axis=-1)
    height_r = np.sqrt(1 - width**2 / 4)
    triangle_area = height_r * width / 2
    #area2 = (width / 4) * np.sqrt(4 - width**2)

    height_w = 2 * triangle_area
    cap_height = np.sqrt(width**2 - height_w**2)

    cap_area = 2 * np.pi * cap_height
    delta_value = 1 / cap_area
    
    def dir_func(dir_vec):
        """
        dir_vec : ndarray of shape (num_directions, 3)
        input of the directional function
        """
        assert dir_vec.ndim == 2
        assert dir_vec.shape[1] == 3
        assert np.isclose(np.linalg.norm(dir_vec, axis=-1), 1).all()
        distance = np.linalg.norm(dir_vec - direction[None,:], axis=-1)
        val = delta_value * (distance < width).astype(float)
        return val
    return dir_func



def test_planewave_integral_of_dirac_coeff_function_returns_sound_pressure():
    """To get a really low error the dirac function must be very narrow. But then the monte
    carlo integration will need a LOT of samples. So this test will be very very slow if it
    is to be reliable. But with the default settings, it is easy to see that the two
    functions are very similar.
    """
    rng = np.random.default_rng()
    wave_dir = mc.uniform_random_on_sphere(1, rng)

    freq = 400
    c = 343
    wave_num = 2 * np.pi * freq / c
    exp_center = rng.uniform(-1, 1, size=(1,3))
    pos_eval = rng.uniform(-1, 1, size=(1,3)) #np.array([[0,1,0]])

    pw_coeff_func = _approx_dirac_planewave_coeffs(wave_dir, 1e-2)
    int_value = pw.plane_wave_integral(pw_coeff_func, pos_eval, exp_center, wave_num, rng, int(1e8))
    pw_value = pw.plane_wave(pos_eval - exp_center, wave_dir, wave_num) 

    diff = int_value - pw_value
    mse = np.mean(np.abs(diff)**2)
    assert mse < 1e-2
    # num_pos = 32
    # posx = np.linspace(-1, 1, num_pos)
    # posy = np.linspace(-1, 1, num_pos)
    # (posx, posy) = np.meshgrid(posx, posy)
    # pos = np.stack((posx.flatten(), posy.flatten(), np.zeros(num_pos**2)), axis=-1)
    


def test_planewave_microphone_model_of_dirac_planewave_returns_directionality_function():
    """ Consider (7) in Brunnström et al 2024. If the sound field is a plane wave, the beta function is a
    dirac function. Then the result of (7) should be the directivity function for the plane wave angle. 

    This is quite slow since we use a monte-carlo integration with a ton of samples. This is because 
    the delta function requires many samples to give a decent approximation of the value. 
    """

    rng = np.random.default_rng(12345)
    freq = 200
    c = 343
    wave_num = 2 * np.pi * freq / c

    mic_angle = mc.uniform_random_on_sphere(1, rng)
    directionality = pw.linear_directivity_function(0.5, mic_angle) #omni_directivity_function() 

    num_tries = 5
    error = np.zeros((num_tries,))
    for i in range(num_tries):
        pw_angle = mc.uniform_random_on_sphere(1, rng)
        dirac_func = _approx_dirac_planewave_coeffs(pw_angle, 1e-2)
        mic_val = pw.apply_measurement(dirac_func, directionality, rng, int(1e8))

        directionality_val = directionality(pw_angle)
        error[i] = np.mean(np.abs(mic_val - directionality_val))

    assert np.mean(error) < 1e-2




def tst_planewave_microphone_model_and_spherical_harmonics_are_identical_for_plane_wave_soundfield():
    """Don't remember what this is supposed to do, so could probably be removed. 
    
    """
    rng = np.random.default_rng()
    max_order = 1
    num_pos = 10000
    pos_mic = np.zeros((1,3))
    exp_center = np.zeros((1,3))
    #pos = np.concatenate((rng.uniform(low = -1, high=1, size=(num_pos,2)), np.zeros((num_pos, 1))), axis=-1)
    pos = rng.uniform(low = -1, high=1, size=(num_pos,3))
    freq = rng.uniform(low=100, high=1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343
    
    planewave_direction = mc.uniform_random_on_sphere(1, rng)
    rad, angles = utils.cart2spherical(planewave_direction)
    orders, degrees = sph.shd_num_degrees_vector(max_order)
    neg_degrees = -degrees


    # harmonic coefficients for a plane wave
    const_factor = (-1.0)**(degrees) * (-1j)**orders * np.sqrt(4 * np.pi)
    sph_harm = special.sph_harm(neg_degrees, orders, angles[...,0], angles[...,1])
    shd_coeffs = const_factor * sph_harm
    shd_coeffs = shd_coeffs[None,:]
    #shd_plane_wave = np.squeeze(sph.reconstruct_pressure(shd_coeffs, pos, np.zeros((1, 3)), wave_num))

    # plane wave sound pressure
    #plane_wave = np.squeeze(pw.plane_wave(pos, planewave_direction, wave_num))

    mic_direction = mc.uniform_random_on_sphere(1, rng)
    dir_coeffs = sph.directivity_linear(0.5, mic_direction)
    #dir_coeffs = sph.directivity_omni()
    assert dir_coeffs.shape == shd_coeffs.shape
    p_shd = np.sum(np.conj(dir_coeffs) * shd_coeffs)
    #p_shd = sph.apply_measurement(shd_coeffs, pos_mic, exp_center, wave_num, dir_coeffs)

    dir_func = pw.linear_directivity_function(0.5, mic_direction)
    #dir_func = pw.omni_directivity_function()
    pw_coeffs = _approx_dirac_planewave_coeffs(planewave_direction, 1e-2)
    p_pw = pw.apply_measurement(pw_coeffs, dir_func, rng, int(1e8))

    pass









# ============= COMPARE MODEL WITH IMAGE SOURCE MODEL =================
def tst_measurement_model_for_analytic_plane_wave_gives_same_value_as_simulated_plane_wave():
    """
    Does not really test anything different than 
    test_measurement_model_with_cardioid_for_analytic_plane_wave_gives_same_value_as_simulated_plane_wave
    so it could probably be removed soon. 
    
    """
    rng = np.random.default_rng()
    sr = 1000
    side_len = 1
    max_order = 16
    
    pw_dir = mc.uniform_random_on_sphere(1, rng)
    mic_dir = mc.uniform_random_on_sphere(1, rng)

    rect = region.Rectangle((side_len, side_len), np.zeros(3), (0.04, 0.04))
    pos_mic = rect.equally_spaced_points()

    p = {}

    pos_mic, p_mic_sim, ir_mic, pos_cardioid, p_cardioid_sim, freqs, sim_info = _generate_plane_wave_with_ism(sr, pw_dir, pos_mic, pos_mic, mic_dir)
    fi = int(p_mic_sim.shape[0] * 0.8)
    p_mic_sim = p_mic_sim[fi,...]
    p_cardioid_sim = p_cardioid_sim[fi,...]

    freq = freqs[fi:fi+1]
    wave_num = 2 * np.pi * freq / sim_info.c
    p_mic_analytic, exp_center, gain_adjustment = pw.find_matching_plane_wave(p_mic_sim, freq, pw_dir, pos_mic, sim_info.c)
    shd_coeffs = gain_adjustment * pw.shd_coeffs_for_planewave(pw_dir, max_order)

    #dir_coeffs_omni = sph.directivity_linear(0, pw_dir)
    #p_omni_shd = sph.apply_measurement(shd_coeffs, pos_mic, exp_center, wave_num, dir_coeffs_omni)

    dir_coeffs_cardioid = sph.directivity_linear(0.5, mic_dir)
    p["cardioid_shd"] = sph.apply_measurement(shd_coeffs, pos_mic, exp_center, wave_num, dir_coeffs_cardioid)

    p["cardioid_sim"] = p_cardioid_sim


    #p_dir_shd = sph.apply_measurement(shd_coeffs, pos_mic, exp_center, wave_num, dir_coeffs)

    p["omni_sim"] = p_mic_sim
    plm.image_scatter_freq_response(p, freq, pos_mic, dot_size = 60)
    plt.show()

    plm.image_scatter_freq_response({"difference" : p["cardioid_shd"] - p["cardioid_sim"]}, freq, pos_mic, dot_size = 60)
    plt.show()


    #assert np.mean(np.abs(p_dir_shd - p_mic_sim)**2) / np.mean(np.abs(p_mic_sim)**2) < 1e-5


def test_measurement_model_with_cardioid_for_analytic_plane_wave_gives_same_value_as_simulated_plane_wave():
    """
    
    As the test test_matched_analytic_plane_wave_is_very_similar_to_simulated_plane_wave shows, the data
    from the simulator and the analytic representation only matches up to an MSE of around 1e-6. Therefore the
    directional measurement model will not match better than that. 
    """
    rng = np.random.default_rng(123456)
    sr = 1000
    max_order = 24
    num_rotations = 30
    
    pw_dir = mc.uniform_random_on_sphere(1, rng)

    pos_mic = rng.uniform(-0.5, 0.5, size=(1,3)) # if chosen too far out (compared to exp_center and max_order) then the shd representation will not be good
    pos_dir = np.ones((num_rotations,1)) * pos_mic

    rect = region.Cuboid((1, 1, 1), np.zeros(3), (0.1, 0.1, 0.1), rng=rng)
    #rect = region.Rectangle((1, 1), np.zeros(3), (0.04, 0.04))
    #pos_omni = rect.equally_spaced_points()
    pos_omni = rect.sample_points(512)

    #mic_angle = np.linspace(0, 2*np.pi, num_rotations, endpoint=False)
    #mic_dir = utils.spherical2cart(np.ones(num_rotations), np.stack((mic_angle, np.pi/2 * np.ones(num_rotations)),axis=-1))
    mic_dir = mc.uniform_random_on_sphere(num_rotations, rng)

    p = {}

    pos_omni, p_omni_sim, ir_mic, pos_cardioid, p_cardioid_sim, freqs, sim_info = _generate_plane_wave_with_ism(sr, pw_dir, pos_omni, pos_dir, mic_dir)
    fi = int(p_omni_sim.shape[0] * 0.8)
    p_omni_sim = p_omni_sim[fi,...]
    p_cardioid_sim = p_cardioid_sim[fi,...]
    p["cardioid_sim"] = p_cardioid_sim

    freq = freqs[fi:fi+1]
    wave_num = 2 * np.pi * freq / sim_info.c
    p_mic_analytic, exp_center, gain_adjustment = pw.find_matching_plane_wave(p_omni_sim, freq, pw_dir, pos_omni, sim_info.c)
    shd_coeffs = gain_adjustment * pw.shd_coeffs_for_planewave(pw_dir, max_order)

    p_omni_shd = sph.apply_measurement_omni(shd_coeffs, pos_omni, exp_center, wave_num)
    mse_plane_wave = np.mean(np.abs(p_omni_shd - p_omni_sim)**2) / np.mean(np.abs(p_omni_sim)**2)
    #plm.image_scatter_freq_response(p, freq, pos_omni, dot_size = 60)
    #plt.show()
    assert mse_plane_wave < 1e-5, "The data used to check the model is not correct"

    dir_coeffs_cardioid = sph.directivity_linear(0.5, mic_dir)[None,...]
    p["cardioid_shd"] = sph.apply_measurement(shd_coeffs, pos_dir, exp_center, wave_num, dir_coeffs_cardioid)

    mse = np.mean(np.abs(p["cardioid_shd"] - p["cardioid_sim"])**2) / np.mean(np.abs(p["cardioid_sim"])**2)

    for name in p:
        p[name] = np.squeeze(p[name])

    # fig, axes = plt.subplots(1, 3, figsize=(10, 10))
    # for name, val in p.items():
    #     axes[0].plot(mic_angle, np.real(val), label=name)
    #     axes[1].plot(mic_angle, np.imag(val), label=name)
    #     axes[2].plot(mic_angle, np.abs(val), label=name)
    # axes[0].set_title("Real part")
    # axes[1].set_title("Imaginary part")
    # axes[2].set_title("Magnitude")
    # axes[0].legend()
    # for ax in axes:
    #     aspplot.set_basic_plot_look(ax)
    #     ylim = ax.get_ylim()
    #     ylim_max = np.max(np.abs(ylim))
    #     ax.set_ylim(-ylim_max, ylim_max)
    # fig.suptitle(f"Total MSE: {mse}")
    # plt.show()

    assert mse < 1e-5, f"MSE: {mse}"



def test_measurement_model_with_cardioid_for_multiple_analytic_plane_waves_gives_same_value_as_simulated_plane_wave():
    """
    This is meant to test whether the cardioid model breaks down when there are multiple plane waves, as it breaks
    down for spherically spreading waves.
    
    As the test test_matched_analytic_plane_wave_is_very_similar_to_simulated_plane_wave shows, the data
    from the simulator and the analytic representation only matches up to an MSE of around 1e-6. Therefore the
    directional measurement model will not match better than that. 
    """
    rng = np.random.default_rng()
    sr = 1000
    max_order = 14
    num_rotations = 30
    num_plane_waves = 5
    
    pw_dir = mc.uniform_random_on_sphere(num_plane_waves, rng)

    pos_mic = rng.uniform(-0.5, 0.5, size=(1,3)) # if chosen too far out (compared to exp_center and max_order) then the shd representation will not be good
    pos_dir = np.ones((num_rotations,1)) * pos_mic

    rect = region.Cuboid((1, 1, 1), np.zeros(3), (0.1, 0.1, 0.1), rng=rng)
    #rect = region.Rectangle((1, 1), np.zeros(3), (0.04, 0.04))
    #pos_omni = rect.equally_spaced_points()
    pos_omni = rect.sample_points(128)

    #mic_angle = np.linspace(0, 2*np.pi, num_rotations, endpoint=False)
    #mic_dir = utils.spherical2cart(np.ones(num_rotations), np.stack((mic_angle, np.pi/2 * np.ones(num_rotations)),axis=-1))
    mic_dir = mc.uniform_random_on_sphere(num_rotations, rng)

    p_cardioid = {}
    p_omni = {}
    shd = {}

    for i in range(num_plane_waves):
        pos_omni, p_omni_sim, ir_mic, pos_cardioid, p_cardioid_sim, freqs, sim_info = _generate_plane_wave_with_ism(sr, pw_dir[i:i+1,:], pos_omni, pos_dir, mic_dir, rng.uniform(300, 900))
        fi = int(p_omni_sim.shape[0] * 0.8)
        p_omni_sim = p_omni_sim[fi,...]
        p_cardioid_sim = p_cardioid_sim[fi,...]
        if not "cardioid_sim" in p_cardioid:
            p_cardioid["cardioid_sim"] = np.zeros_like(p_cardioid_sim)
        if not "omni_sim" in p_omni:
            p_omni["omni_sim"] = np.zeros_like(p_omni_sim)
        p_cardioid["cardioid_sim"] += p_cardioid_sim
        p_omni["omni_sim"] += p_omni_sim

        freq = freqs[fi:fi+1]
        wave_num = 2 * np.pi * freq / sim_info.c
        p_mic_analytic, exp_center, gain_adjustment = pw.find_matching_plane_wave(p_omni_sim, freq, pw_dir[i:i+1,:], pos_omni, sim_info.c)
        shd_coeffs = gain_adjustment * pw.shd_coeffs_for_planewave(pw_dir[i:i+1,:], max_order)
        new_exp_center = np.zeros((1,3))
        shd_coeffs = sph.translate_shd_coeffs(shd_coeffs, new_exp_center-exp_center, wave_num, max_order)

        if not "shd_coeffs" in shd:
            shd["shd_coeffs"] = np.zeros_like(shd_coeffs)
        shd["shd_coeffs"] += shd_coeffs
        
    p_omni_shd = sph.apply_measurement_omni(shd["shd_coeffs"], pos_omni, new_exp_center, wave_num)
    mse_plane_wave = np.mean(np.abs(p_omni_shd - p_omni["omni_sim"])**2) / np.mean(np.abs(p_omni["omni_sim"])**2)
    #plm.image_scatter_freq_response(p, freq, pos_omni, dot_size = 60)
    #plt.show()
    assert mse_plane_wave < 1e-5, "The data used to check the model is not correct"

    dir_coeffs_cardioid = sph.directivity_linear(0.5, mic_dir)[None,...]
    p_cardioid["cardioid_shd"] = sph.apply_measurement(shd["shd_coeffs"], pos_dir, new_exp_center, wave_num, dir_coeffs_cardioid)

    mse = np.mean(np.abs(p_cardioid["cardioid_shd"] - p_cardioid["cardioid_sim"])**2) / np.mean(np.abs(p_cardioid["cardioid_sim"])**2)

    for name in p_cardioid:
        p_cardioid[name] = np.squeeze(p_cardioid[name])

    # fig, axes = plt.subplots(1, 3, figsize=(10, 10))
    # for name, val in p_cardioid.items():
    #     axes[0].plot(mic_angle, np.real(val), label=name)
    #     axes[1].plot(mic_angle, np.imag(val), label=name)
    #     axes[2].plot(mic_angle, np.abs(val), label=name)
    # axes[0].set_title("Real part")
    # axes[1].set_title("Imaginary part")
    # axes[2].set_title("Magnitude")
    # axes[0].legend()
    # for ax in axes:
    #     aspplot.set_basic_plot_look(ax)
    #     ylim = ax.get_ylim()
    #     ylim_max = np.max(np.abs(ylim))
    #     ax.set_ylim(-ylim_max, ylim_max)
    # fig.suptitle(f"Total MSE: {mse}")
    # plt.show()

    assert mse < 1e-5, f"MSE: {mse}"







def test_matched_analytic_plane_wave_is_very_similar_to_simulated_plane_wave():
    rng = np.random.default_rng()
    sr = 2000
    side_len = 1

    pw_dir = mc.uniform_random_on_sphere(1, rng)

    #rect = region.Rectangle((side_len, side_len), np.zeros(3), (0.04, 0.04))
    #pos_mic = rect.equally_spaced_points()
    rect = region.Cuboid((side_len, side_len, side_len), np.zeros(3), rng=rng)
    pos_mic = rect.sample_points(1000)
    

    pos_mic, p_mic_sim, ir_mic_sim, _, _, freqs, sim_info = _generate_plane_wave_with_ism(sr, pw_dir, pos_mic)
    fi = int(p_mic_sim.shape[0] * 0.8)
    p_mic_sim = p_mic_sim[fi,...]

    freq = freqs[fi:fi+1]
    p_mic_analytic, _, _ = pw.find_matching_plane_wave(p_mic_sim, freq, pw_dir, pos_mic, sim_info.c)

    #plm.image_scatter_freq_response({"sim" : p_mic_sim, "analytic" : p_mic_analytic}, freq, pos_mic, dot_size = 60)
    #plt.show()

    mse = np.mean(np.abs(p_mic_sim - p_mic_analytic)**2) / np.mean(np.abs(p_mic_sim)**2)
    ##plt.plot(np.squeeze(ir_mic_sim).T)
    #plt.title(mse)
    #plt.show()
    assert mse < 1e-5

def test_matched_analytic_plane_wave_is_identical_to_original_analytic_plane_wave():
    rng = np.random.default_rng()
    freq = np.ones(1) * 400
    c = 343
    wave_num = 2 * np.pi * freq / c
    side_len = 1

    pw_dir = mc.uniform_random_on_sphere(1, rng)
    rect = region.Cuboid((side_len, side_len, side_len), np.zeros(3), rng=rng)
    pos_mic = rect.sample_points(1000)

    p_mic_orig = np.squeeze(pw.plane_wave(pos_mic, pw_dir, wave_num), axis=-1)
    p_mic_analytic, _, _ = pw.find_matching_plane_wave(p_mic_orig, freq, pw_dir, pos_mic, c)
    assert np.mean(np.abs(p_mic_orig - p_mic_analytic)**2) / np.mean(np.abs(p_mic_orig)**2) < 1e-20

def test_matched_analytic_plane_wave_pressure_and_reconstructed_analytic_plane_wave_are_identical():
    rng = np.random.default_rng()
    sr = 1000
    pw_dir = mc.uniform_random_on_sphere(1, rng)

    pos_mic, p_mic_sim, _, _, _, freqs, sim_info = _generate_plane_wave_with_ism(sr, pw_dir)
    fi = int(p_mic_sim.shape[0] * 0.8)
    p_mic_sim = p_mic_sim[fi,...]

    freq = freqs[fi:fi+1]
    wave_num = 2 * np.pi * freq / sim_info.c
    p_mic_matched, exp_center, gain_adjustment = pw.find_matching_plane_wave(p_mic_sim, freq, pw_dir, pos_mic, sim_info.c)

    pw_reconstructed = np.squeeze(pw.plane_wave(pos_mic, pw_dir, wave_num, exp_center), axis=-1)
    pw_reconstructed = pw_reconstructed * gain_adjustment

    assert np.mean(np.abs(pw_reconstructed - p_mic_matched)**2) / np.mean(np.abs(p_mic_matched)**2) < 1e-20
    
def test_matched_analytic_plane_wave_reconstruction_and_spherical_harmonic_reconstruction_are_identical():
    rng = np.random.default_rng()
    sr = 1000
    max_order = 30
    pw_dir = mc.uniform_random_on_sphere(1, rng)

    side_len = 1
    rect = region.Cuboid((side_len, side_len, side_len), np.zeros(3), rng=rng)
    pos_mic = rect.sample_points(1000)

    pos_mic, p_mic_sim, _, _, _, freqs, sim_info = _generate_plane_wave_with_ism(sr, pw_dir, pos_mic)
    fi = int(p_mic_sim.shape[0] * 0.8)
    p_mic_sim = p_mic_sim[fi,...]

    freq = freqs[fi:fi+1]
    wave_num = 2 * np.pi * freq / sim_info.c
    p_mic_matched, exp_center, gain_adjustment = pw.find_matching_plane_wave(p_mic_sim, freq, pw_dir, pos_mic, sim_info.c)

    shd_coeffs = gain_adjustment * pw.shd_coeffs_for_planewave(pw_dir, max_order)
    p_mic_shd = sph.reconstruct_pressure(shd_coeffs, pos_mic, exp_center, wave_num)
    assert np.mean(np.abs(p_mic_shd - p_mic_matched)**2) / np.mean(np.abs(p_mic_matched)**2) < 1e-15



def _generate_plane_wave_with_ism(sr, pw_dir, pos_mic=None, pos_dir=None, mic_dir=None, src_distance=1000):
    rng = np.random.default_rng(10)
    side_len = 0.5
    num_mic = 40

    assert np.allclose(np.linalg.norm(pw_dir), 1)

    #pos_mic = np.zeros((num_mic, 3))
    if pos_mic is None:
        pos_mic = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))

    if pos_dir is None:
        pos_dir = np.zeros((1,3))
        mic_dir = np.array([[1,0,0]])
    
    #src_distance = 1000
    pos_src = pw_dir * src_distance 

    setup = _get_default_simulator_setup(sr)
    prop_delay = np.linalg.norm(pos_src) / setup.sim_info.c
    extra_delay_sec = 0.1
    rir_len = coreutils.next_power_of_two((1.5 * prop_delay + extra_delay_sec) * sr)
    extra_delay_samples = int(extra_delay_sec * sr)

    setup.sim_info.room_size = [3 * src_distance, 3*src_distance, 3*src_distance]
    setup.sim_info.room_center = [0, 0, 0]
    setup.sim_info.rt60 =  0
    setup.sim_info.max_room_ir_length = rir_len
    setup.sim_info.sim_buffer = rir_len
    setup.sim_info.extra_delay = extra_delay_samples

    setup.add_mics("omni", pos_mic)
    if mic_dir.shape[0] == 1:
        mic_dir = np.ones((pos_dir.shape[0], 1)) * mic_dir
    setup.add_mics("cardioid", pos_dir, directivity_type=pos_dir.shape[0]*["cardioid"], directivity_dir=mic_dir)
    setup.add_controllable_source("src", pos_src)
    sim = setup.create_simulator()

    #sim.arrays.paths["src"]["omni"] = 10000 * sim.arrays.paths["src"]["omni"]
    #sim.arrays.paths["src"]["cardioid"] = 10000 * sim.arrays.paths["src"]["cardioid"]

    num_freqs = setup.sim_info.max_room_ir_length
    #fpaths, freqs = sim.arrays.get_freq_paths(num_freqs, sr)
    fpaths, freqs = _get_freq_paths_correct_time_convention(sim.arrays, num_freqs, sr)

    return sim.arrays["omni"].pos, fpaths["src"]["omni"][...,0], sim.arrays.paths["src"]["omni"], \
            sim.arrays["cardioid"].pos, fpaths["src"]["cardioid"][...,0], freqs, sim.sim_info



def _get_freq_paths_correct_time_convention(arrays, num_freqs, samplerate):
    """Get the frequency domain response of the paths between all sources and microphones
    
    Parameters
    ----------
    arrays : ArrayCollection
        The array collection to get the frequency paths from
    num_freqs : int
        The number of frequencies to calculate the response for. The number refers to the total FFT
        length, so the number of real frequencies is num_freqs // 2 + 1, which is the number of 
        frequency bins in the output. 
    
    Returns
    -------
    freq_paths : dict of dicts of ndarrays
        freq_paths[src_name][mic_name] is a complex ndarray of shape (num_real_freqs, num_mic, num_src)
        representing the frequency response between a source and a microphone array
    freqs : ndarray of shape (num_real_freqs,)
        The real frequencies in Hz of the response

    Notes
    -----
    The frequencies returned are compatible with get_real_freqs of the package aspcol, as well as the
    output of np.fft.rfftfreq. 

    num_freqs can be safely chosen as a larger number than the number of samples in the impulse response,
    as the FFT will zero-pad the signal to the desired length. But if num_freqs is smaller, then the
    signal will be truncated.
    """
    if num_freqs % 2 == 1:
        raise NotImplementedError("Odd number of frequencies not supported yet")
    
    freqs = (samplerate / (num_freqs)) * np.arange(num_freqs // 2 + 1)
    num_real_freqs = ft.get_real_freqs(num_freqs, samplerate).shape[-1]

    fpaths = {}
    for src, mic, path in arrays.iter_paths():
        fpaths.setdefault(src.name, {})
        freq_domain_path = ft.rfft(path, n=num_freqs)
        fpaths[src.name][mic.name] = np.moveaxis(freq_domain_path, 1,2)
    return fpaths, freqs




def _get_default_simulator_setup(sr):
    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 =  0.25
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 60
    setup.sim_info.plot_output = "none"
    return setup





import numpy as np
import pathlib
import hypothesis as hyp
#from hypothesis import given
import hypothesis.strategies as st
import pytest
import itertools as it
from sympy.physics.wigner import wigner_3j, gaunt

import scipy.special as special

import aspcore.montecarlo as mc
import aspcore.fouriertransform as ft

import aspcol.sphericalharmonics as sph
import aspcol.utilities as utils


import aspcol.kernelinterpolation as ki

import aspcol.planewaves as pw
import aspcol.movingmicrophone as mm

import wigner
import wigners

from aspsim.simulator import SimulatorSetup

import matplotlib.pyplot as plt

import plot_methods as plm






# ========= GAUNT COEFFICIENTS =========
def _gaunt_coef_ueno(l1, m1, l2, m2, l3):
    """Gaunt coefficients

    Taken directly from https://github.com/sh01k/MeshRIR/blob/main/example/sf_func.py
    """
    m3 = -m1 - m2
    l = int((l1 + l2 + l3) / 2)

    t1 = l2 - m1 - l3
    t2 = l1 + m2 - l3
    t3 = l1 + l2 - l3
    t4 = l1 - m1
    t5 = l2 + m2

    tmin = max([0, max([t1, t2])])
    tmax = min([t3, min([t4, t5])])

    t = np.arange(tmin, tmax+1)
    gl_tbl = np.array(special.gammaln(np.arange(1, l1+l2+l3+3)))
    G = np.sum( (-1.)**t * np.exp( -np.sum( gl_tbl[np.array([t, t-t1, t-t2, t3-t, t4-t, t5-t])] )  \
                                  +np.sum( gl_tbl[np.array([l1+l2-l3, l1-l2+l3, -l1+l2+l3, l])] ) \
                                  -np.sum( gl_tbl[np.array([l1+l2+l3+1, l-l1, l-l2, l-l3])] ) \
                                  +np.sum( gl_tbl[np.array([l1+m1, l1-m1, l2+m2, l2-m2, l3+m3, l3-m3])] ) * 0.5 ) ) \
        * (-1.)**( l + l1 - l2 - m3) * np.sqrt( (2*l1+1) * (2*l2+1) * (2*l3+1) / (4*np.pi) )
    return G


def gaunt_sympy(l1, l2, l3, m1, m2, m3):
    f1 = np.sqrt((2*l1+1)*(2*l2+1)*(2*l3+1)/(4*np.pi))
    f2 = wigner_3j(l1,l2,l3,m1,m2,m3)
    f3 = wigner_3j(l1,l2,l3,0,0,0) 
    return f1*f2*f3


def test_gaunt_coef_dev():
    max_order = 3

    all_params = np.array([i for i in sph._all_gaunt_parameters(max_order)])
    #all_params = all_params[7:8]

    val = []
    val_ref = []
    for i in range(all_params.shape[0]):
        args = all_params[i]
        val.append(sph.gaunt_coefficient(*args))
        val_ref.append(_gaunt_coef_ueno(*args))

    val_ref = np.array(val_ref)
    val = np.array(val)
    assert np.allclose(val, val_ref)


def test_gaunt_coef():
    max_order = 3
    all_params = np.array([i for i in sph._all_gaunt_parameters(max_order)])
    val = sph.gaunt_coefficient(all_params[:, 0], all_params[:, 1], all_params[:, 2], all_params[:, 3], all_params[:, 4])

    val_ref = []
    for i in range(all_params.shape[0]):
        args = all_params[i]
        val_ref.append(_gaunt_coef_ueno(*args))
    val_ref = np.array(val_ref)
    assert np.allclose(val, val_ref)


def test_gaunt_coef_lookup():
    max_order = 6
    all_params = np.array([i for i in sph._all_gaunt_parameters(max_order)])

    tbl = sph.gaunt_coef_lookup(max_order)
    val = tbl[all_params[:, 0], all_params[:, 1], all_params[:, 2], all_params[:, 3], all_params[:, 4]]

    val_ref = []
    for i in range(all_params.shape[0]):
        args = all_params[i]
        val_ref.append(_gaunt_coef_ueno(*args))
    val_ref = np.array(val_ref)
    assert np.allclose(val, val_ref)


# ============ GAUNT AND WIGNER TESTS ============

def _linearity_formula_lower_limit(l1, l2, m1, m2):
    """ Lower limit of the sum in the linearity formula for the Gaunt coefficient
    described in (3.74) in Multiple scattering: Interaction of time-harmonic waves with N obstacles.

    func(l1, l2, m1, m2) is equivalent to func(l1, l1, m2, m1)
    """
    if np.abs(l1 - l2) >= np.abs(m1 + m2):
        return np.abs(l1 - l2)
    elif np.abs(l1 - l2) < np.abs(m1 + m2) and ((l1 + l2 + np.abs(m1 + m2)) % 2 == 0):
        return np.abs(m1 + m2)
    elif np.abs(l1 - l2) < np.abs(m1 + m2) and ((l1 + l2 + np.abs(m1 + m2)) % 2 == 1):
        return np.abs(m1 + m2) + 1
    else:
        raise ValueError("Something went wrong")

def test_relationship_between_gaunt_and_triple_harmonic_integral():
    max_order = 5
    all_params = np.array([i for i in sph._all_gaunt_parameters(max_order)])
    num_param_sets = all_params.shape[0]
    assert num_param_sets > 0 

    for i in range(num_param_sets):
        (l1, m1, l2, m2, l3) = all_params[i]
        val1 = sph.triple_harmonic_integral(l1, l2, l3, m1, m2, -m1-m2)
        val2 = (-1)**(m1+m2) * sph.gaunt_coefficient(l1, m1, l2, m2, l3)
        assert np.allclose(val1, val2)


def test_triple_harmonic_integral_against_sympy():
    max_order = 3
    all_params = np.array([i for i in sph._all_gaunt_parameters(max_order)])
    num_param_sets = all_params.shape[0]
    assert num_param_sets > 0 

    for i in range(num_param_sets):
        (l1, m1, l2, m2, l3) = all_params[i]
        val1 = sph.triple_harmonic_integral(l1, l2, l3, m1, m2, -m1-m2)
        val2 = float(gaunt(l1, l2, l3, m1, m2, -m1-m2).n(16))
        assert np.allclose(val1, val2)

def test_all_wigner_3j_implementations_are_equal():
    max_order = 8
    all_params = np.array([i for i in sph._all_gaunt_parameters(max_order)])
    num_param_sets = all_params.shape[0]
    assert num_param_sets > 0 

    for i in range(num_param_sets):
        (l1, m1, l2, m2, l3) = all_params[i]

        m2_min, m2_max, wigner_val = wigner.wigner_3jm(l1,l2,l3,m1)
        val1 = wigner_val[m2 - int(m2_min)]

        val2 = wigners.wigner_3j(l1,l2,l3,m1,m2,-m1-m2)
        val3 = float(wigner_3j(l1,l2,l3,m1,m2,-m1-m2).n(16))
        assert np.allclose(val1, val2)
        assert np.allclose(val1, val3)



# =============== SPHERICAL HARMONIC TESTS ========================
def test_linearisation_formula_for_spherical_harmonics():
    """
    Equation (3.70) in Multiple scattering: Interaction of time-harmonic waves with N obstacles
    Relates the Gaunt coefficient to the product of two spherical harmonics. A simpler expression
    for the same thing is given between (3.74) and (3.75) in the same book.
    """
    rng = np.random.default_rng()
    direction = rng.uniform(-1, 1, size = (1,3))
    radius, angles = utils.cart2spherical(direction)

    max_order = 5
    for l1, l2 in it.product(range(max_order+1), range(max_order+1)):
        for m1, m2, in it.product(range(-l1, l1+1), range(-l2, l2+1)):
            q0 = _linearity_formula_lower_limit(l1, l2, m1, m2)
            Q = (l1 + l2 - q0) / 2
            assert utils.is_integer(Q)
            Q = int(Q)
            
            sum_val = 0
            for l3 in range(Q+1):
                g = sph.gaunt_coefficient(l1, m1, l2, m2, q0 + 2*l3)
                sum_val += g * special.sph_harm(m1+m2, q0 + 2*l3, angles[0,0], angles[0,1])

            comparison_val = special.sph_harm(m1, l1, angles[0,0], angles[0,1]) * special.sph_harm(m2, l2, angles[0,0], angles[0,1])
            assert np.allclose(sum_val, comparison_val)

def test_basis_function_equals_integral_of_spherical_harmonics_and_plane_waves():
    """This is the relationship that is used to define the directivity coefficients, so it
    must hold for the directional microphone model to make sense. 

    This is the Funk-Hecke formula, multiplied by sqrt(4 pi) on both sides. 
    Eq. 3.43 in Multiple scattering: interaction of time-harmonic waves with N obstacles. 

    basis_nu^mu(r) = frac{i^nu}{sqrt{4 pi}} int_S^2 Y_nu^mu(d_hat) exp(-ikr^T d_hat)    ds(d_hat)
    """
    rng = np.random.default_rng()
    max_order = 7
    pos = rng.uniform(low = -1, high=1, size=(1,3))
    freq = rng.uniform(low=100, high=1000)
    wave_num = freq / 343

    orders, degrees = sph.shd_num_degrees_vector(max_order)
    basis_val = np.squeeze(sph.shd_basis(pos, orders, degrees, wave_num))

    def calc_integral(num_samples):
        sphere_samples = rng.normal(loc = 0, scale = 1, size=(num_samples, 3))
        sphere_samples = sphere_samples / np.linalg.norm(sphere_samples, axis=-1)[:,None]
        rad, angles = utils.cart2spherical(sphere_samples)

        const_factor = 1j**orders / np.sqrt(4 * np.pi)
        sph_harm = special.sph_harm(degrees[:,None], orders[:,None], angles[...,0], angles[...,1])
        plane_wave = np.exp(-1j * wave_num * np.sum(pos * sphere_samples, axis=-1))
        
        sphere_area = 4 * np.pi
        int_value = const_factor * sphere_area * np.mean(sph_harm * plane_wave[None,:], axis=-1) # multiply by area?
        return int_value
    
    num_samples_all = 10 ** np.arange(2, 7)
    int_values = [calc_integral(ns) for ns in num_samples_all]
    mse = [np.mean(np.abs(basis_val - int_val)**2) for int_val in int_values]

    # UNCOMMENT TO SEE ERROR EVOLUTION
    # fig, ax = plt.subplots(1,1, figsize=(8,6))
    # ax.plot(np.log10(num_samples_all), 10 * np.log10(mse))
    # ax.set_title("Error between basis function and integral")
    # ax.set_xlabel("Number of samples [log10]")
    # ax.set_ylabel("Mean square error [dB]")
    # plt.show()

    assert all([mse[i] >= mse[i+1] for i in range(len(mse)-1)]) # integral should be more exact with more samples
    assert mse[-1] < 1e-5 #integral should be very close to basis function

def test_plane_wave_with_positive_exponential_expressed_in_spherical_harmonics():
    """This is the Rayleigh expansion formula
    Eq. 3.69 in Multiple scattering: interaction of time-harmonic waves with N obstacles.
    It expressed a single plane wave in terms of spherical harmonics.

    exp(ik r^T d) = sum_{nu, mu} varphi_{nu}^{mu}(r) sqrt(4 pi) i^{nu} conj{Y_{nu}^{mu}(d)}
    
    It can be viewed as a regular spherical decomposition, with coefficients defined as
    alpha_{nu}^{mu} = sqrt(4 pi) i^{nu} conj{Y_{nu}^{mu}(d)}
    and with an expansion centered around the origin.
    """
    rng = np.random.default_rng()
    max_order = 40
    num_pos = 10000
    # pos = np.concatenate((rng.uniform(low = -1, high=1, size=(num_pos,2)), np.zeros((num_pos, 1))), axis=-1)
    pos = rng.uniform(low = -1, high=1, size=(num_pos,3))
    freq = rng.uniform(low=100, high=1000)
    wave_num = 2 * np.pi * freq / 343
    
    planewave_direction = mc.uniform_random_on_sphere(1, rng)
    rad, angles = utils.cart2spherical(planewave_direction)
    orders, degrees = sph.shd_num_degrees_vector(max_order)

    # Calculate harmonic coefficients
    const_factor = 1j**orders * np.sqrt(4 * np.pi)
    sph_harm = special.sph_harm(degrees, orders, angles[...,0], angles[...,1])
    shd_coeffs = const_factor * np.conj(sph_harm)
    shd_plane_wave = np.squeeze(sph.reconstruct_pressure(shd_coeffs, pos, np.zeros((1, 3)), wave_num))

    plane_wave = np.squeeze(pw.plane_wave(pos, planewave_direction, wave_num))
    plane_wave = np.conj(plane_wave) # conjugate because we need exp(ikr^T d)
    
    # UNCOMMENT TO SEE PLANE WAVE AND THE APPROXIMATE PLANE WAVE FROM SPHERICAL HARMONICS
    #plm.image_scatter_freq_response(plane_wave, np.array([freq]), pos, dot_size=10)
    #plm.image_scatter_freq_response(sph_plane_wave, np.array([freq]), pos, dot_size=10)
    #plt.show()
    mse = np.mean(np.abs(plane_wave - shd_plane_wave)**2) / np.mean(np.abs(plane_wave)**2)
    assert mse < 1e-10

def test_plane_wave_with_negative_exponential_expressed_in_spherical_harmonics():
    """This is the Rayleigh expansion formula, but for plane waves of the form exp(-ikr^T d)
    It is a slight reformulation of Eq. 3.69 in Multiple scattering: interaction of time-harmonic waves with N obstacles.
    Using identity for conjugation of spherical harmonics to retain the form of the expression as a
    sperical harmonic decomposition. 

    exp(-ik r^T d) = sum_{nu, mu}  sqrt(4 pi) (-i)^{nu} (-1)^{mu} Y_{nu}^{-mu}(d) varphi_{nu}^{mu}(r)
    
    It can be viewed as a regular spherical decomposition, with coefficients defined as
    alpha_{nu}^{mu} = sqrt(4 pi) (-1)^{mu} (-i)^{nu} Y_{nu}^{-mu}(d)
    and with an expansion centered around the origin.
    """
    rng = np.random.default_rng()
    max_order = 40
    num_pos = 10000
    #pos = np.concatenate((rng.uniform(low = -1, high=1, size=(num_pos,2)), np.zeros((num_pos, 1))), axis=-1)
    pos = rng.uniform(low = -1, high=1, size=(num_pos,3))
    freq = rng.uniform(low=100, high=1000)
    wave_num = 2 * np.pi * freq / 343
    
    planewave_direction = mc.uniform_random_on_sphere(1, rng)
    rad, angles = utils.cart2spherical(planewave_direction)
    orders, degrees = sph.shd_num_degrees_vector(max_order)
    neg_degrees = -degrees

    # Calculate harmonic coefficients
    const_factor = (-1.0)**(degrees) * (-1j)**orders * np.sqrt(4 * np.pi)
    sph_harm = special.sph_harm(neg_degrees, orders, angles[...,0], angles[...,1])
    shd_coeffs = const_factor * sph_harm
    shd_plane_wave = np.squeeze(sph.reconstruct_pressure(shd_coeffs, pos, np.zeros((1, 3)), wave_num))

    plane_wave = np.squeeze(pw.plane_wave(pos, planewave_direction, wave_num))
    
    # UNCOMMENT TO SEE PLANE WAVE AND THE APPROXIMATE PLANE WAVE FROM SPHERICAL HARMONICS
    # plm.image_scatter_freq_response(plane_wave, np.array([freq]), pos, dot_size=10)
    # plm.image_scatter_freq_response(shd_plane_wave, np.array([freq]), pos, dot_size=10)
    # plt.show()
    mse = np.mean(np.abs(plane_wave - shd_plane_wave)**2) / np.mean(np.abs(plane_wave)**2)
    assert mse < 1e-10


# ============ TRANSLATION OPERATOR ============
def show_translation_operator_addition_property():
    """
    The translation operator should satisfy 
    T(r + r') = T(r) T(r')
    """
    rng = np.random.default_rng(123456)
    pos1 = rng.normal(size = (1,3))
    pos2 = rng.normal(size = (1,3))
    pos_added = pos1 + pos2
    wave_num = np.array([2 * np.pi * 1000 / 343])
    max_order_input = 2
    max_order_mid = 20
    max_order_output = 2

    T1 = sph.translation_operator(pos1, wave_num, max_order_input, max_order_mid)
    T2 = sph.translation_operator(pos2, wave_num, max_order_mid, max_order_output)
    Tadded = sph.translation_operator(pos_added, wave_num, max_order_input, max_order_output)
    T = T2 @ T1

    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    clr = axes[0].matshow(np.squeeze(np.abs(T)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(T)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(T)))
    plt.colorbar(clr, ax = axes[2])

    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    clr = axes[0].matshow(np.squeeze(np.abs(Tadded)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(Tadded)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(Tadded)))
    plt.colorbar(clr, ax = axes[2])
    plt.show()



# ================ SHD BASIS TRANSLATION ================

def test_basis_translation_addition_property_as_function_of_max_order_mid():
    """The translation operator should satisfy T(r + r') = T(r) T(r')
    This test shows that as we increase the truncation order of the
    internal dimension of T(r) T(r'), i.e. the one dimension that disappears
    after multiplication, the MSE should go towards zero. 
    """
    rng = np.random.default_rng(123456)
    pos1 = rng.normal(size = (1,3))
    pos2 = rng.normal(size = (1,3))
    pos_added = pos1 + pos2
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    max_order_input = 2
    max_order_output = 2
    Tadded = sph.basis_translation_3_92(pos_added, wave_num, max_order_input, max_order_output)
    norm_factor = np.mean(np.abs(Tadded)**2)

    max_order_mid = [i for i in range(2, 21, 2)]
    mse = []
    for mom in max_order_mid:
        T1 = sph.basis_translation_3_92(pos1, wave_num, max_order_input, mom)
        T2 = sph.basis_translation_3_92(pos2, wave_num, mom, max_order_output)
        T = T2 @ T1

        mse.append(10 * np.log10(np.mean(np.abs(T - Tadded)**2) / norm_factor))
    # Uncomment to see how the mse evolve with increasing max_order_mid
    # plt.plot(max_order_mid, mse)
    # plt.xlabel("max_order_mid")
    # plt.ylabel("MSE (dB)")
    # plt.show()
    assert all([mse[i] >= mse[i+1] for i in range(len(mse)-1)])


def test_basis_translation_with_negative_argument_is_hermitian_basis_translation():
    """The translation operator should satisfy T(-r) = T(r)^H
    """
    rng = np.random.default_rng()
    pos = rng.normal(size = (1,3))
    wave_num = np.array([2 * np.pi * 1000 / 343])
    max_order_input = 5
    max_order_output = 5

    T = np.squeeze(sph.basis_translation_3_92(pos, wave_num, max_order_input, max_order_output), axis=(0,1))
    T2 = np.squeeze(sph.basis_translation_3_92(-pos, wave_num, max_order_input, max_order_output), axis=(0,1))

    # plt.matshow(np.abs(T))
    # plt.colorbar()
    # plt.matshow(np.abs(T2))
    # plt.colorbar()
    # plt.show()

    assert np.allclose(T, T2.conj().T)

def test_basis_translation_with_zero_arguments_is_identity():
    """Identity given as (i) on page 88 in 
    P. Martin's Multiple scattering: Interaction of time-harmonic waves with N obstacles
    
    The translation operator should satisfy hat{S}(0) = I
    """
    pos = np.zeros((1,3))
    wave_num = np.array([2 * np.pi * 1000 / 343])
    max_order_input = 5
    max_order_output = 5
    T = np.squeeze(sph.basis_translation_3_92(pos, wave_num, max_order_input, max_order_output))
    assert np.allclose(T, np.eye(T.shape[0]))

def test_basis_translation_with_input_order_zero_gives_basis_function():
    """Identity given as (ii) on page 88 in 
    P. Martin's Multiple scattering: Interaction of time-harmonic waves with N obstacles
    """
    rng = np.random.default_rng()
    pos = rng.normal(size=(1,3))
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    max_order_output = 5
    T = np.squeeze(sph.basis_translation_3_92(pos, wave_num, 0, max_order_output))

    orders, degrees = sph.shd_num_degrees_vector(max_order_output)
    basis_vals = np.squeeze(sph.shd_basis(pos, orders, degrees, wave_num))
    assert np.allclose(basis_vals, T)

def test_basis_translation_with_output_order_zero_gives_conjugated_basis_function():
    """Identity given as (iii) on page 88 in 
    P. Martin's Multiple scattering: Interaction of time-harmonic waves with N obstacles
    """
    rng = np.random.default_rng()
    pos = rng.normal(size=(1,3))
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    max_order_input = 5
    T = np.squeeze(sph.basis_translation_3_92(pos, wave_num, max_order_input, 0))

    orders, degrees = sph.shd_num_degrees_vector(max_order_input)
    basis_vals = np.conj(np.squeeze(sph.shd_basis(pos, orders, degrees, wave_num)))
    for i, nu in enumerate(orders):
        basis_vals[i] = (-1)**nu * basis_vals[i]
    assert np.allclose(basis_vals, T)


def test_basis_translation_Martin_identity_3_93():
    """Identity given as (3.93) on page 92 in P. Martin,
    hat{S}_{n nu}^{m mu} = (-1)^{n+nu} hat{S}_{n nu}^{m mu}.conj()
    """
    rng = np.random.default_rng()
    pos = rng.normal(size=(1,3))
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    max_order = 5
    T = np.squeeze(sph.basis_translation_3_92(pos, wave_num, max_order, max_order), axis=(0,1))
    orders, degrees = sph.shd_num_degrees_vector(max_order)
    for i, (n, m) in enumerate(zip(orders, degrees)):
        for j, (nu, mu) in enumerate(zip(orders, degrees)):
            assert np.allclose(T[j,i], (-1.0)**(n + nu) * np.conj(T[i,j]))

def test_basis_translation_Martin_identity_3_94():
    """Identity given as (3.94) on page 92 in P. Martin,
    hat{S}_{n nu}^{-m -mu} = (-1)^{m+mu} hat{S}_{n nu}^{m mu}.conj()
    """
    rng = np.random.default_rng()
    pos = rng.normal(size=(1,3))
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    max_order = 5
    T = np.squeeze(sph.basis_translation_3_92(pos, wave_num, max_order, max_order), axis=(0,1))
    orders, degrees = sph.shd_num_degrees_vector(max_order)
    for i, (n, m) in enumerate(zip(orders, degrees)):
        for j, (nu, mu) in enumerate(zip(orders, degrees)):
            i2 = _find_index_of_order_degree(n, -m)
            j2 = _find_index_of_order_degree(nu, -mu)
            assert np.allclose(T[i2,j2], (-1.0)**(m + mu) * np.conj(T[i,j]))

def _find_index_of_order_degree(order, degree):
    orders, degrees = sph.shd_num_degrees_vector(order)
    idx = np.where((orders == order) & (degrees == degree))[0]
    return idx[0]


def test_basis_translation_3_80_is_equivalent_to_basis_translation_3_92():
    rng = np.random.default_rng()
    pos = rng.normal(size=(4,3))
    freq = rng.uniform(100, 1000, size=(2,))
    wave_num = 2 * np.pi * freq / 343
    max_order_input = rng.integers(1, 10)
    max_order_output = rng.integers(1, 10)
    T1 = sph.basis_translation_3_80(pos, wave_num, max_order_input, max_order_output)
    T2 = sph.basis_translation_3_92(pos, wave_num, max_order_input, max_order_output)
    assert np.allclose(T1, T2)




# ============== TRANSLATION OF SHD COEFFICIENTS ==============
def test_translation_operator_with_negative_argument_is_hermitian():
    """The translation operator should satisfy T(-r) = T(r)^H
    """
    rng = np.random.default_rng()
    pos = rng.normal(size = (1,3))
    wave_num = np.array([2 * np.pi * 1000 / 343])
    max_order_input = 5
    max_order_output = 5

    T = np.squeeze(sph.translation_operator(pos, wave_num, max_order_input, max_order_output), axis=(0,1))
    T2 = np.squeeze(sph.translation_operator(-pos, wave_num, max_order_input, max_order_output), axis=(0,1))
    assert np.allclose(T, T2.conj().T)

def test_translation_operator_addition_property_as_function_of_max_order_mid():
    """The translation operator should satisfy T(r + r') = T(r) T(r')
    This test shows that as we increase the truncation order of the
    internal dimension of T(r) T(r'), i.e. the one dimension that disappears
    after multiplication, the MSE should go towards zero. 
    """
    rng = np.random.default_rng(123456)
    pos1 = rng.normal(size = (1,3))
    pos2 = rng.normal(size = (1,3))
    pos_added = pos1 + pos2
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    max_order_input = 3
    max_order_output = 3
    Tadded = sph.translation_operator(pos_added, wave_num, max_order_input, max_order_output)
    norm_factor = np.mean(np.abs(Tadded)**2)

    max_order_mid = [i for i in range(2, 21, 2)]
    mse = []
    for mom in max_order_mid:
        T1 = sph.translation_operator(pos1, wave_num, max_order_input, mom)
        T2 = sph.translation_operator(pos2, wave_num, mom, max_order_output)
        T = T2 @ T1

        mse.append(10 * np.log10(np.mean(np.abs(T - Tadded)**2) / norm_factor))
    # Uncomment to see how the mse evolve with increasing max_order_mid
    # plt.plot(max_order_mid, mse)
    # plt.xlabel("max_order_mid")
    # plt.ylabel("MSE (dB)")
    # plt.show()
    assert all([mse[i] >= mse[i+1] for i in range(len(mse)-1)])

def test_translation_operator_addition_property():
    """The translation operator should satisfy T(r + r') = T(r) T(r') 
    """
    rng = np.random.default_rng(123456)
    pos1 = rng.normal(size = (1,3))
    pos2 = rng.normal(size = (1,3))
    pos_added = pos1 + pos2
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    max_order_input = 6
    max_order_output = 6
    Tadded = sph.translation_operator(pos_added, wave_num, max_order_input, max_order_output)

    max_order_mid = 18
    T1 = sph.translation_operator(pos1, wave_num, max_order_input, max_order_mid)
    T2 = sph.translation_operator(pos2, wave_num, max_order_mid, max_order_output)
    T = T2 @ T1

    mse = 10 * np.log10(np.mean(np.abs(T - Tadded)**2) / np.mean(np.abs(Tadded)**2))
    assert mse < -100


# def test_translation_operator_against_uenos_implementation():
#     """
#     It looks like uenos implemenation differs. Not sure how
#     to determine whether that's a mistake or not. 
#     """
#     rng = np.random.default_rng(21345654)
#     pos = rng.normal(size = (1,3))
#     freq = rng.uniform(100, 1000)
#     wave_num = np.array([2 * np.pi * freq / 343])
#     max_order_input = 5
#     max_order_output = 5

#     T_ueno = sph._translation_ueno(pos, wave_num, max_order_input, max_order_output)[0,0,:,:]
#     T = sph.translation_operator(pos, wave_num, max_order_input, max_order_output)[0,0,:,:]
#     # UNCOMMENT TO VIEW DIFFERENCES
#     # fig, axes = plt.subplots(1, 3, figsize=(14,4))
#     # axes[0].set_title("our")
#     # clr = axes[0].matshow(np.squeeze(np.abs(T)))
#     # plt.colorbar(clr, ax = axes[0])
#     # clr = axes[1].matshow(np.squeeze(np.real(T)))
#     # plt.colorbar(clr, ax = axes[1])
#     # clr = axes[2].matshow(np.squeeze(np.imag(T)))
#     # plt.colorbar(clr, ax = axes[2])

#     # fig, axes = plt.subplots(1, 3, figsize=(14,4))
#     # axes[0].set_title("Ueno")
#     # clr = axes[0].matshow(np.squeeze(np.abs(T_ueno)))
#     # plt.colorbar(clr, ax = axes[0])
#     # clr = axes[1].matshow(np.squeeze(np.real(T_ueno)))
#     # plt.colorbar(clr, ax = axes[1])
#     # clr = axes[2].matshow(np.squeeze(np.imag(T_ueno)))
#     # plt.colorbar(clr, ax = axes[2])

#     # fig, axes = plt.subplots(1, 3, figsize=(14,4))
#     # axes[0].set_title("Difference")
#     # clr = axes[0].matshow(np.squeeze(np.abs(T_ueno-T)))
#     # plt.colorbar(clr, ax = axes[0])
#     # clr = axes[1].matshow(np.squeeze(np.real(T_ueno-T)))
#     # plt.colorbar(clr, ax = axes[1])
#     # clr = axes[2].matshow(np.squeeze(np.imag(T_ueno-T)))
#     # plt.colorbar(clr, ax = axes[2])
#     # plt.show()
#     assert np.allclose(T, T_ueno)


def test_translation_operator_against_jax_compiled_implementation():
    rng = np.random.default_rng(21345654)
    pos = rng.normal(size = (3,3))
    freq = rng.uniform(100, 1000, size=(2,))
    wave_num = 2 * np.pi * freq / 343
    max_order_input = 2
    max_order_output = 2

    gaunt = mm.calculate_gaunt_set(max_order_input, max_order_output)
    T_jax = mm.translation_operator(pos, wave_num, max_order_input, max_order_output, gaunt)
    T = sph.translation_operator(pos, wave_num, max_order_input, max_order_output)

    T_jax = T_jax[0,0,:,:]
    T = T[0,0,:,:]
    # UNCOMMENT TO VIEW DIFFERENCES
    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    axes[0].set_title("Numpy")
    clr = axes[0].matshow(np.squeeze(np.abs(T)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(T)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(T)))
    plt.colorbar(clr, ax = axes[2])

    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    axes[0].set_title("Jax")
    clr = axes[0].matshow(np.squeeze(np.abs(T_jax)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(T_jax)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(T_jax)))
    plt.colorbar(clr, ax = axes[2])

    fig, axes = plt.subplots(1, 3, figsize=(14,4))
    axes[0].set_title("Difference")
    clr = axes[0].matshow(np.squeeze(np.abs(T_jax-T)))
    plt.colorbar(clr, ax = axes[0])
    clr = axes[1].matshow(np.squeeze(np.real(T_jax-T)))
    plt.colorbar(clr, ax = axes[1])
    clr = axes[2].matshow(np.squeeze(np.imag(T_jax-T)))
    plt.colorbar(clr, ax = axes[2])
    plt.show()
    assert np.allclose(T, T_jax)


















def test_translation_of_shd_coeffs_does_not_affect_reconstructed_function():
    """The reconstructed function should be unaffected by translation
    """
    rng = np.random.default_rng()
    side_len = 0.3
    
    num_side = 20
    pos_eval = np.meshgrid(np.linspace(-side_len/2, side_len/2, num_side), np.linspace(-side_len/2, side_len/2, num_side), np.zeros(1,))
    pos_eval = np.concatenate(pos_eval, axis=-1)
    pos_eval = np.reshape(pos_eval, (-1,3))

    exp_center = rng.uniform(low=-0.1, high=0.1, size=(1,3))
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343
    max_order = 5
    max_order_output = max_order + 10
    shd_coeffs = _random_shd_coeffs(max_order, 1, rng)

    p = sph.reconstruct_pressure(shd_coeffs, pos_eval, exp_center, wave_num)
    pos_translate = rng.uniform(low = -0.2, high = 0.2, size=(1,3))
    new_exp_center = exp_center + pos_translate

    shd_coeffs_translated = sph.translate_shd_coeffs(shd_coeffs, pos_translate, wave_num, max_order_output)
    p_translated = sph.reconstruct_pressure(shd_coeffs_translated, pos_eval, new_exp_center, wave_num)
    p_translated_nocompensation = sph.reconstruct_pressure(shd_coeffs_translated, pos_eval, exp_center, wave_num)

    diff = p - p_translated

    # UNCOMMENT TO VIEW DIFFERENCES
    # p_all = {"original": p, "translated": p_translated, "difference": diff, "nocompensation" : p_translated_nocompensation, "diff nocompensation" : p - p_translated_nocompensation,}
    # plm.image_scatter_freq_response(p_all, freq, pos_eval)
    # plt.show()
    assert np.mean(np.abs(diff)) < 1e-8


def test_translation_of_shd_coeffs_gives_same_values_as_directly_estimated_shd_coeffs():
    sr = 1000
    reg = 1e-7
    max_order = 3
    center1 = np.zeros((1,3))
    center2 = np.ones((1,3)) * 0.2

    pos1, p1, freqs, sim_info = _generate_soundfield_data_omni(sr, center1)
    pos2, p2, freqs, sim_info = _generate_soundfield_data_omni(sr, center2)
    wave_num = 2 * np.pi * freqs / sim_info.c

    shd_coeffs1 = sph.inf_dimensional_shd_omni(p1, pos1, center1, max_order, wave_num, reg)
    shd_coeffs2 = sph.inf_dimensional_shd_omni(p2, pos2, center2, max_order, wave_num, reg)

    shd_coeffs2_est = sph.translate_shd_coeffs(shd_coeffs1, center2-center1, wave_num, max_order)

    fig, axes = plt.subplots(1, 4, figsize=(20,6))
    axes[0].plot(10*np.log10(np.mean(np.abs(shd_coeffs2 - np.squeeze(shd_coeffs2_est))**2, axis=-1)))
    axes[0].set_ylabel("Mean square error (dB)")
    axes[0].set_xlabel("Frequency (Hz)")

    axes[1].plot(10*np.log10(np.mean(np.abs(shd_coeffs2 - np.squeeze(shd_coeffs2_est))**2, axis=0) / np.mean(np.abs(shd_coeffs2)**2, axis=0)))
    axes[1].set_ylabel("Mean square error (dB)")
    axes[1].set_xlabel("Frequency (Hz)")

    axes[2].plot(np.real(shd_coeffs2_est[:,0]), label="estimated")
    axes[2].plot(np.real(shd_coeffs2[:,0]), label="recorded")
    axes[2].set_title("Real part")
    # axes[2].plot(np.imag(shd_coeffs2_est), label="estimated")
    # axes[2].plot(np.imag(shd_coeffs2), label="recorded") 
    # axes[2].set_title("Imag part")
    axes[3].plot(np.abs(shd_coeffs2_est[:,0]), label="estimated")
    axes[3].plot(np.abs(shd_coeffs2[:,0]), label="recorded")
    axes[3].set_title("Magnitude")
    for ax in axes:
        ax.legend()
    plt.show()
    


def test_translation_of_shd_coeffs_is_numerically_stable_for_small_arguments():
    """The reconstructed function should be unaffected by translation
    """
    rng = np.random.default_rng(1234567)
    side_len = 0.3
    
    num_side = 20
    pos_eval = np.meshgrid(np.linspace(-side_len/2, side_len/2, num_side), np.linspace(-side_len/2, side_len/2, num_side), np.zeros(1,))
    pos_eval = np.concatenate(pos_eval, axis=-1)
    pos_eval = np.reshape(pos_eval, (-1,3))

    exp_center = np.zeros((1,3)) #rng.uniform(low=-0.1, high=0.1, size=(1,3))
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343
    max_order = 1
    max_order_output = max_order + 10
    shd_coeffs = _random_shd_coeffs(max_order, 1, rng)

    p = sph.reconstruct_pressure(shd_coeffs, pos_eval, exp_center, wave_num)
    pos_translate = 1e-9 * np.array([[0,1,0]]) #rng.uniform(low = -0.2, high = 0.2, size=(1,3))
    new_exp_center = exp_center + pos_translate

    shd_coeffs_translated = sph.translate_shd_coeffs(shd_coeffs, pos_translate, wave_num, max_order_output)
    p_translated = sph.reconstruct_pressure(shd_coeffs_translated, pos_eval, new_exp_center, wave_num)
    p_translated_nocompensation = sph.reconstruct_pressure(shd_coeffs_translated, pos_eval, exp_center, wave_num)

    diff = p - p_translated

    # UNCOMMENT TO VIEW DIFFERENCES
    # p_all = {"original": p, "translated": p_translated, "difference": diff, "nocompensation" : p_translated_nocompensation, "diff nocompensation" : p - p_translated_nocompensation,}
    # plm.image_scatter_freq_response(p_all, freq, pos_eval)
    # plt.show()
    assert np.mean(np.abs(diff)) < 1e-8






 #====================== OTHER SHD FUNCTIONS ========================

def test_translated_inner_product_for_zero_order_is_sinc_function():
    rng = np.random.default_rng()
    num_pos = 10
    pos1 = rng.normal(size = (num_pos,3))
    pos2 = rng.normal(size = (num_pos,3))
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343
    val = sph.translated_inner_product(pos1, pos2, sph.directivity_omni(), sph.directivity_omni(), wave_num)
    val2 = ki.kernel_helmholtz_3d(pos1, pos2, wave_num)
    assert np.allclose(val, val2)

def test_measurement_conj_omni_times_measurement_omni_is_kernel_matrix():
    """
    Xi Xi^H = Psi where Psi is the matrix given in Ueno. For omni microphones as
    in this case, Psi equals K the kernel matrix from Ueno's kernel interpolation paper. 
    """
    rng = np.random.default_rng()
    num_mic1 = 3
    num_mic2 = 4
    pos1 = rng.normal(size=(num_mic1, 3))
    pos2 = rng.normal(size=(num_mic2, 3))
    exp_center = rng.normal(size = (1, 3)) * 0
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343

    input_vec = rng.normal(size=(1, num_mic2)) + 1j * rng.normal(size=(1, num_mic2))

    psi = ki.kernel_helmholtz_3d(pos1, pos2, wave_num)
    vec_ip = np.squeeze(psi @ input_vec[...,None], axis=-1)

    mse = []
    max_inner_order = 40
    for i in np.arange(5, max_inner_order, 5):
        shd_meas = sph.apply_measurement_conj_omni(input_vec, pos2, exp_center, i, wave_num)
        vec_meas = sph.apply_measurement_omni(shd_meas, pos1, exp_center, wave_num)
        mse.append(10 * np.log10(np.mean(np.abs(vec_meas - vec_ip)**2)))
    #plt.plot(np.arange(2, 35, 2), mse)
    #plt.show()
    assert all([mse[i] >= mse[i+1] or mse[i] < -200 for i in range(len(mse)-1)])
    assert np.allclose(vec_ip, vec_meas)

def test_measurement_conj_omni_times_measurement_omni_is_translated_inner_product():
    """
    Xi Xi^H = Psi where Psi is the matrix given in Ueno. For omni microphones as
    in this case, Psi equals K the kernel matrix from Ueno's kernel interpolation paper. 
    """
    rng = np.random.default_rng()
    num_mic1 = 3
    num_mic2 = 4
    pos1 = rng.normal(size=(num_mic1, 3))
    pos2 = rng.normal(size=(num_mic2, 3))
    exp_center = rng.normal(size = (1, 3)) * 0
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343

    input_vec = rng.normal(size=(1, num_mic2)) + 1j * rng.normal(size=(1, num_mic2))

    dir_coeffs1 = sph.directivity_omni()
    dir_coeffs2 = sph.directivity_omni()
    psi = sph.translated_inner_product(pos1, pos2, dir_coeffs1, dir_coeffs2, wave_num)
    vec_ip = np.squeeze(psi @ input_vec[...,None], axis=-1)

    mse = []
    max_inner_order = 40
    for i in np.arange(5, max_inner_order, 5):
        shd_meas = sph.apply_measurement_conj_omni(input_vec, pos2, exp_center, i, wave_num)
        vec_meas = sph.apply_measurement_omni(shd_meas, pos1, exp_center, wave_num)
        mse.append(10 * np.log10(np.mean(np.abs(vec_meas - vec_ip)**2)))
    #plt.plot(np.arange(2, 35, 2), mse)
    #plt.show()
    assert all([mse[i] >= mse[i+1] or mse[i] < -100 for i in range(len(mse)-1)])
    assert np.allclose(vec_ip, vec_meas)

def test_measurement_operator_times_conjugate_measurement_operator_equals_translated_inner_product():
    rng = np.random.default_rng()
    num_mic1 = 2
    num_mic2 = 3
    pos1 = rng.normal(size=(num_mic1, 3))
    pos2 = rng.normal(size=(num_mic2, 3))
    exp_center = rng.normal(size = (1, 3)) * 0
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343

    max_order1 = 2
    max_order2 = 2

    input_vec = rng.normal(size=(1, num_mic2))
    dir_coeffs1 = _random_shd_coeffs(max_order1, 1, rng)
    dir_coeffs2 = _random_shd_coeffs(max_order2, 1, rng)

    psi = sph.translated_inner_product(pos1, pos2, dir_coeffs1, dir_coeffs2, wave_num)
    vec_ip = np.squeeze(psi @ input_vec[...,None], axis=-1)

    mse = []
    max_inner_order = 41
    for inner_order in np.arange(10, max_inner_order, 10):
        shd_meas = sph.apply_measurement_conj(input_vec, pos2, exp_center, inner_order, wave_num, dir_coeffs2)
        vec_meas = sph.apply_measurement(shd_meas, pos1, exp_center, wave_num, dir_coeffs1)
        mse.append(10 * np.log10(np.mean(np.abs(vec_meas - vec_ip)**2 / np.mean(np.abs(vec_ip)**2))))

    #plt.plot(np.arange(10, max_inner_order, 10), mse)
    #plt.show()
    assert all([mse[i] >= mse[i+1] or mse[i] < -200 for i in range(len(mse)-1)])
    assert np.allclose(vec_ip, vec_meas)

# ============= SHD COEFFICIENTS AND RECONSTRUCTIONS =================
def test_good_reconstruction_using_shd_estimation():
    sr = 1000
    ds_freq = 20
    exp_center = np.zeros((1,3))
    pos_mic, rirs, freqs, sim_info = _generate_soundfield_data_omni(sr, exp_center)
    wave_num = 2 * np.pi * freqs / sim_info.c
    max_order = 9

    rirs = np.ascontiguousarray(rirs[::ds_freq,...])
    wave_num = np.ascontiguousarray(wave_num[::ds_freq])
    freqs = np.ascontiguousarray(freqs[::ds_freq])

    shd_coeffs = sph.inf_dimensional_shd_omni(rirs, pos_mic, exp_center, max_order, wave_num, 1e-8)

    p_est = sph.reconstruct_pressure(shd_coeffs, pos_mic, exp_center, wave_num)
    mse = 10 * np.log10(np.mean(np.abs(p_est - rirs)**2, axis=-1))

    # plt.plot(freqs, mse)
    # plt.show()
    assert np.mean(mse) < -50


def test_measurement_operator_omni_is_same_as_reconstruct_pressure():
    """Xi omni should be the same as the pressure reconstructed from the SHD coefficients
    """
    rng = np.random.default_rng()

    max_order = 10
    num_pos = 1
    freq = rng.uniform(100, 1000)
    wave_num = np.array([2 * np.pi * freq / 343])
    exp_center = rng.normal(size = (1, 3))

    shd_coeffs = _random_shd_coeffs(max_order, 1, rng)
    pos = rng.normal(size=(num_pos,3))
    
    p = sph.reconstruct_pressure(shd_coeffs, pos, exp_center, wave_num)
    p_xi = sph.apply_measurement_omni(shd_coeffs, pos, exp_center, wave_num)
    assert np.allclose(p, p_xi)


def test_measurement_operator_conj_omni_is_adjoint_to_measurement_operator_omni():
    rng = np.random.default_rng()
    num_mic = 10
    max_order = 5
    pos = rng.normal(size=(num_mic, 3))
    exp_center = rng.normal(size = (1, 3))
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343

    vec = rng.normal(size=(1, num_mic))
    shd_coeffs = _random_shd_coeffs(max_order, 1, rng)

    new_vec = sph.apply_measurement_omni(shd_coeffs, pos, exp_center, wave_num)
    val = np.sum(new_vec * np.conj(vec))

    new_shd_sequence = sph.apply_measurement_conj_omni(vec, pos, exp_center, max_order, wave_num)
    val_from_conj = np.sum(shd_coeffs * np.conj(new_shd_sequence))

    assert np.allclose(val, val_from_conj)

def test_measurement_operator_conj_is_adjoint_to_measurement_operator():
    rng = np.random.default_rng()
    num_mic = 10
    max_order = 5
    pos = rng.normal(size=(num_mic, 3))
    exp_center = rng.normal(size = (1, 3))
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343

    vec = rng.normal(size=(1, num_mic))
    shd_coeffs = _random_shd_coeffs(max_order, 1, rng)
    dir_coeffs = _random_shd_coeffs(max_order, 1, rng)

    new_vec = sph.apply_measurement(shd_coeffs, pos, exp_center, wave_num, dir_coeffs)
    val = np.sum(new_vec * np.conj(vec))

    new_shd_sequence = sph.apply_measurement_conj(vec, pos, exp_center, max_order, wave_num, dir_coeffs)
    val_from_conj = np.sum(shd_coeffs * np.conj(new_shd_sequence))

    assert np.allclose(val, val_from_conj)

def test_measurement_operator_conj_omni_is_hermitian_transpose_of_measurement_operator_omni():
    rng = np.random.default_rng()
    num_mic = 10
    max_order = 5
    pos = rng.normal(size=(num_mic, 3))
    exp_center = rng.normal(size = (1, 3))
    freq = rng.uniform(100, 1000, size=(1,))
    wave_num = 2 * np.pi * freq / 343

    meas = sph.measurement_omni(pos, exp_center, max_order, wave_num)
    conj_meas = sph.measurement_conj_omni(pos, exp_center, max_order, wave_num)
    assert np.allclose(np.moveaxis(meas.conj(), -1, -2), conj_meas)









# ================== HELPER FUNCTIONS ==========================



def _random_shd_coeffs(max_order, num_freqs, rng=None):
    if rng is None:
        rng = np.random.default_rng()

    shd_coeffs = rng.normal(size=(num_freqs, sph.shd_num_coeffs(max_order))) + \
        1j * rng.normal(size=(num_freqs, sph.shd_num_coeffs(max_order)))
    return shd_coeffs




def _generate_soundfield_data_omni(sr, exp_center = np.array([[0,0,0]])):
    rng = np.random.default_rng(10)
    side_len = 0.2
    num_mic = 100

    #pos_mic = np.zeros((num_mic, 3))
    pos_mic = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))
    pos_mic += exp_center

    pos_src = np.array([[3,0.05,-0.05]])

    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 =  0.25
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 100
    setup.sim_info.plot_output = "none"

    setup.add_mics("omni", pos_mic)
    setup.add_controllable_source("src", pos_src)

    sim = setup.create_simulator()

    num_freqs = setup.sim_info.max_room_ir_length 
    fpaths, freqs = _get_freq_paths_correct_time_convention(sim.arrays, num_freqs, sr)

    return sim.arrays["omni"].pos, fpaths["src"]["omni"][...,0], freqs, sim.sim_info




def _generate_soundfield_data_dir_center(sr, directivity_dir, rt60 = 0.25):
    rng = np.random.default_rng(10)
    side_len = 0.1
    num_mic = 30

    pos_mic = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))

    pos_src = np.array([[0,1.5,-0.05]])

    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 = rt60
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 40
    setup.sim_info.plot_output = "none"

    setup.add_mics("omni", pos_mic)
    setup.add_mics("dir", np.zeros((1,3)), directivity_type=["cardioid"], directivity_dir=directivity_dir)
    setup.add_controllable_source("src", pos_src)

    sim = setup.create_simulator()

    num_freqs = 2048
    fpaths, freqs = _get_freq_paths_correct_time_convention(sim.arrays, num_freqs, sr)

    return sim.arrays["omni"].pos, fpaths["src"]["omni"][...,0], fpaths["src"]["dir"][...,0], freqs, sim.sim_info



def _generate_soundfield_data_dir_surrounded_by_sources(sr, directivity_dir, num_src, rt60 = 0.25):
    rng = np.random.default_rng(10)
    side_len = 0.4
    num_mic = 30

    pos_mic = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))

    src_angles = np.linspace(0, 2*np.pi, num_src, endpoint=False)
    pos_src = utils.spherical2cart(np.ones((num_src,)), np.stack((src_angles, np.pi/2*np.ones((num_src,))), axis=-1))

    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr
    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 = rt60
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 40
    setup.sim_info.plot_output = "none"

    setup.add_mics("omni", pos_mic)
    setup.add_mics("dir", np.zeros((1,3)), directivity_type=["cardioid"], directivity_dir=directivity_dir)
    setup.add_controllable_source("src", pos_src)

    sim = setup.create_simulator()

    num_freqs = 512
    fpaths, freqs = _get_freq_paths_correct_time_convention(sim.arrays, num_freqs, sr)

    return sim.arrays["omni"].pos, sim.arrays["src"].pos, sim.arrays["dir"].pos, fpaths["src"]["omni"], fpaths["src"]["dir"], freqs, sim.sim_info


def _generate_soundfield_data_dir_and_omni(sr, directivity_dir, rt60 = 0.25):
    rng = np.random.default_rng(10)
    side_len = 1
    num_mic = 50

    pos_mic = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))
    pos_mic_dir = np.copy(pos_mic)

    pos_src = np.array([[0,1.5,-0.05]])

    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 = rt60
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 40
    setup.sim_info.plot_output = "none"

    setup.add_mics("omni", pos_mic)
    dir_type = num_mic*["cardioid"]
    dir_dir = np.tile(directivity_dir, (num_mic,1))
    setup.add_mics("dir", pos_mic_dir, directivity_type=dir_type, directivity_dir=dir_dir)
    setup.add_controllable_source("src", pos_src)

    sim = setup.create_simulator()

    num_freqs = 512
    fpaths, freqs = _get_freq_paths_correct_time_convention(sim.arrays, num_freqs, sr)

    return sim.arrays["omni"].pos, fpaths["src"]["omni"][...,0], fpaths["src"]["dir"][...,0], freqs, sim.sim_info


def __temp_gen_data():
    rng = np.random.default_rng(10)
    side_len = 1
    num_mic = 40
    num_eval = 800
    sr = 1000

    #pos_mic = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))
    pos_mic = np.concatenate((rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 2)), np.zeros((num_mic,1))), axis=-1)
    pos_eval = np.concatenate((rng.uniform(low=-side_len/2, high=side_len/2, size=(num_eval, 2)), np.zeros((num_eval,1))), axis=-1)
    #pos_eval = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))

    pos_src = np.array([[3,0.05,-0.05]])

    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 =  0.1
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 100
    setup.sim_info.plot_output = "none"

    setup.add_mics("omni", pos_mic)
    setup.add_mics("eval", pos_eval)
    setup.add_controllable_source("src", pos_src)

    sim = setup.create_simulator()

    num_freqs = setup.sim_info.max_room_ir_length

    fpaths, freqs = _get_freq_paths_correct_time_convention(sim.arrays, num_freqs, sr)

    return sim.arrays["omni"].pos, sim.arrays["eval"].pos, fpaths["src"]["omni"][...,0], fpaths["src"]["eval"][...,0], freqs, sim.sim_info


def test_temp_ki():
    import aspcol.soundfieldestimation as sfe
    import aspcol.unittests.plot_methods as plm
    import aspcol.plot as aspplot

    pos_mic, pos_eval, p_mic, p_eval, freqs, sim_info = __temp_gen_data()
    wave_num = 2 * np.pi *freqs / sim_info.c

    
    est = {}
    est["diffuse KRR"] = sfe.est_ki_diffuse_freq(p_mic, pos_mic, pos_eval, wave_num, 1e-5)
    est["nearest neighbour"] = sfe.nearest_neighbour_freq(p_mic, pos_mic, pos_eval)

    est_all = {"diffuse KRR": est["diffuse KRR"], "nearest neighbour": est["nearest neighbour"], "true": p_eval}
    plm.image_scatter_freq_response(est_all, freqs, pos_eval, dot_size = 25)

    fig, ax = plt.subplots(1,1)
    for est_name, p_est in est.items():
        mse = 10 * np.log10(np.mean(np.abs(p_est - p_eval)**2, axis=-1) / np.mean(np.abs(p_eval)**2, axis=-1))
        ax.plot(freqs, mse, label=est_name)
    ax.legend()
    aspplot.set_basic_plot_look(ax)
    plt.show()


def _get_freq_paths_correct_time_convention(arrays, num_freqs, samplerate):
    """Get the frequency domain response of the paths between all sources and microphones
    
    Parameters
    ----------
    arrays : ArrayCollection
        The array collection to get the frequency paths from
    num_freqs : int
        The number of frequencies to calculate the response for. The number refers to the total FFT
        length, so the number of real frequencies is num_freqs // 2 + 1, which is the number of 
        frequency bins in the output. 
    
    Returns
    -------
    freq_paths : dict of dicts of ndarrays
        freq_paths[src_name][mic_name] is a complex ndarray of shape (num_real_freqs, num_mic, num_src)
        representing the frequency response between a source and a microphone array
    freqs : ndarray of shape (num_real_freqs,)
        The real frequencies in Hz of the response

    Notes
    -----
    The frequencies returned are compatible with get_real_freqs of the package aspcol, as well as the
    output of np.fft.rfftfreq. 

    num_freqs can be safely chosen as a larger number than the number of samples in the impulse response,
    as the FFT will zero-pad the signal to the desired length. But if num_freqs is smaller, then the
    signal will be truncated.
    """
    if num_freqs % 2 == 1:
        raise NotImplementedError("Odd number of frequencies not supported yet")
    
    freqs = ft.get_real_freqs(num_freqs, samplerate)

    fpaths = {}
    for src, mic, path in arrays.iter_paths():
        fpaths.setdefault(src.name, {})
        freq_domain_path = ft.rfft(path, n=num_freqs)
        fpaths[src.name][mic.name] = np.moveaxis(freq_domain_path, 1,2)
    return fpaths, freqs


"""Helper functions for various tasks without a clear category


References
----------
"""
import numpy as np
import datetime
import time
from functools import wraps
import numpy as np

def cart2pol(x, y):
    """Transforms the provided cartesian coordinates to polar coordinates
    
    Parameters
    ----------
    x : float or ndarray of shape (num_points,)
        x coordinate
    y : float or ndarray of shape (num_points,)
        y coordinate

    Returns
    -------
    r : float or ndarray of shape (num_points,)
        radius
    angle : float or ndarray of shape (num_points,)
        angle in radians. 0 is the x-direction, pi/2 is the y-direction
    
    """
    r = np.hypot(x, y)
    angle = np.arctan2(y, x)
    return (r, angle)

def pol2cart(r, angle):
    """Tranforms the provided polar coordinates to cartesian coordinates
    
    Parameters
    ----------
    r : float or ndarray of shape (num_points,)
        radius
    angle : float or ndarray of shape (num_points,)
        angle in radians. 0 is the x-direction, pi/2 is the y-direction
    
    Returns
    -------
    x : float or ndarray of shape (num_points,)
        x coordinate
    y : float or ndarray of shape (num_points,)
        y coordinate
    """
    x = r * np.cos(angle)
    y = r * np.sin(angle)
    return (x, y)

def cart2spherical(cart_coord):
    """Transforms the provided cartesian coordinates to spherical coordinates

    Parameters
    ----------
    cart_coord : ndarray of shape (num_points, 3)

    Returns
    -------
    r : ndarray of shape (num_points, 1)
        radius of each point
    angle : ndarray of shape (num_points, 2)
        angle[:,0] is theta, the angle in the xy plane, where 0 is x direction, pi/2 is y direction
        angle[:,1] is phi, the zenith angle, where 0 is z direction, pi is negative z direction
    """
    r = np.linalg.norm(cart_coord, axis=1)
    r_xy = np.linalg.norm(cart_coord[:,:2], axis=1)

    theta = np.arctan2(cart_coord[:,1], cart_coord[:,0])
    phi = np.arctan2(r_xy, cart_coord[:,2])
    angle = np.concatenate((theta[:,None], phi[:,None]), axis=1)
    return (r, angle)

def spherical2cart(r, angle):
    """Transforms the provided spherical coordinates to cartesian coordinates
    
    Parameters
    ----------
    r : ndarray of shape (num_points, 1) or (num_points,)
        radius of each point
    angle : ndarray of shape (num_points, 2)
        the angles in radians
        angle[:,0] is theta, the angle in the xy plane, where 0 is x direction, pi/2 is y direction
        angle[:,1] is phi, the zenith angle, where 0 is z direction, pi is negative z direction
    
    Returns
    -------
    cart_coord : ndarray of shape (num_points, 3)
        the cartesian coordinates
    """
    num_points = r.shape[0]
    cart_coord = np.zeros((num_points,3))
    cart_coord[:,0] = np.squeeze(r) * np.cos(angle[:,0]) * np.sin(angle[:,1])
    cart_coord[:,1] = np.squeeze(r) * np.sin(angle[:,0]) * np.sin(angle[:,1])
    cart_coord[:,2] = np.squeeze(r) * np.cos(angle[:,1])
    return cart_coord


def db2mag(db):
    """Transforms the provided decibel value to magnitude"""
    return 10 ** (db / 20)

def mag2db(amp):
    """Transforms the provided magnitude to decibel"""
    return 20 * np.log10(amp)

def db2pow(db):
    """Transforms the provided decibel value to power"""
    return 10 ** (db / 10)

def pow2db(power):
    """Transforms the provided power to decibel"""
    return 10 * np.log10(power)




def measure_time(name):
    """
    Use as decorator on a function to measure the time the function takes
    use as @measure_time('print_this_as_name')
    """
    def measure_internal(func):
        @wraps(func)
        def _time_it(*args, **kwargs):
            start = int(round(time.time() * 1000))
            try:
                return func(*args, **kwargs)
            finally:
                end_ = int(round(time.time() * 1000)) - start
                print(name + f" execution time: {end_ if end_ > 0 else 0} ms")
        return _time_it
    return measure_internal



def flatten_dict(dict_to_flatten, parent_key="", sep="~"):
    items = []
    for key, value in dict_to_flatten.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, dict):
            items.extend(flatten_dict(value, new_key, sep).items())
        else:
            items.append((new_key, value))
    new_dict = dict(items)
    return new_dict


def restack_dict(dict_to_stack, sep="~"):
    """Only accepts dicts of depth 2.
    All elements must be of that depth."""
    extracted_data = {}
    for multi_key in dict_to_stack.keys():
        key_list = multi_key.split(sep)
        if len(key_list) > 2:
            raise NotImplementedError
        if key_list[0] not in extracted_data:
            extracted_data[key_list[0]] = {}
        extracted_data[key_list[0]][key_list[1]] = dict_to_stack[multi_key]
    return extracted_data



def get_time_string(detailed=False):
    """Returns a string with the current time in the format 'year_month_day_hour_minute'
    
    Parameters
    ----------
    detailed : bool
        If True, seconds and microseconds will be included in the string

    Returns
    -------
    time_str : str
        The time string
    """
    tm = datetime.datetime.now()
    time_str = (
        str(tm.year)
        + "_"
        + str(tm.month).zfill(2)
        + "_"
        + str(tm.day).zfill(2)
        + "_"
        + str(tm.hour).zfill(2)
        + "_"
        + str(tm.minute).zfill(2)
    )
    if detailed:
        time_str += "_" + str(tm.second).zfill(2)
        time_str += "_" + str(tm.microsecond).zfill(2)
    return time_str


def get_unique_folder(prefix, parent_folder, detailed_naming=False):
    """Returns a unique folder name in the parent folder with the prefix and the current time

    The folder name has the form parent_folder / prefix_year_month_day_hour_minute_0. If multiple folders are created
    within the same minute, the number is incremented by 1 for each new folder. 

    Parameters
    ----------
    prefix : str
        The prefix for the folder name
    parent_folder : Path
        The parent folder where the new folder should be created, as a Path object (from pathlib)
    detailed_naming : bool
        If True, the folder name will include seconds and microseconds. 
        If used with multithreading, it is a good idea to set this to True. 
        In that case, uniqueness is not guaranteed, but it reduces the risk of clashes significantly.
    
    Returns
    -------
    folder_name : Path
        The full path to the new folder. The folder is not created by this function.
    """
    file_name = prefix + get_time_string(detailed=detailed_naming)
    file_name += "_0"
    folder_name = parent_folder / file_name
    if folder_name.exists():
        idx = 1
        folder_name_len = len(folder_name.name) - 2
        while folder_name.exists():
            new_name = folder_name.name[:folder_name_len] + "_" + str(idx)
            folder_name = folder_name.parent / new_name
            idx += 1
    return folder_name